######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains the Tool project item class."""
import os
from PySide6.QtCore import QItemSelection, Qt, Slot
from PySide6.QtGui import QAction
from spine_engine.config import TOOL_OUTPUT_DIR
from spine_engine.project_item.project_item_resource import CmdLineArg, LabelArg, make_cmd_line_arg
from spine_engine.utils.helpers import ExecutionDirection, resolve_julia_executable, resolve_python_interpreter
from spine_engine.utils.serialization import deserialize_path, serialize_path
from spinetoolbox.helpers import open_url, select_root_directory, same_path
from spinetoolbox.mvcmodels.file_list_models import FileListModel
from ..commands import UpdateCmdLineArgsCommand, UpdateGroupIdCommand, UpdateRootDirCommand, UpdateResultDirCommand
from ..db_writer_item_base import DBWriterItemBase
from ..models import ToolCommandLineArgsModel
from .commands import (
    UpdateKillCompletedProcesses,
    UpdateLogProcessOutput,
    UpdateToolExecuteInWorkCommand,
    UpdateToolOptionsCommand,
)
from .executable_item import ExecutableItem
from .item_info import ItemInfo
from .output_resources import scan_for_resources
from .utils import find_file, flatten_file_path_duplicates
from .widgets.custom_menus import ToolSpecificationMenu
from .widgets.options_widgets import JuliaOptionsWidget


class Tool(DBWriterItemBase):
    def __init__(
        self,
        name,
        description,
        x,
        y,
        toolbox,
        project,
        specification_name="",
        execute_in_work=True,
        cmd_line_args=None,
        options=None,
        kill_completed_processes=False,
        log_process_output=False,
        group_id=None,
        root_dir="",
    ):
        """
        Args:
            name (str): Object name
            description (str): Object description
            x (float): Initial X coordinate of item icon
            y (float): Initial Y coordinate of item icon
            toolbox (ToolboxUI): QMainWindow instance
            project (SpineToolboxProject): The project this item belongs to
            specification_name (str): Name of this Tool's Tool specification
            execute_in_work (bool): Execute associated Tool specification in work (True) or source directory (False)
            cmd_line_args (list, optional): Tool command line arguments
            options (dict, optional): Misc tool options
            kill_completed_processes (bool): Whether to kill completed persistent processes
            log_process_output (bool): Whether to log process output to a file
            group_id (str, optional): Execution group id
            root_dir (str, optional): Root directory for the Tool Spec's program
        """
        super().__init__(name, description, x, y, project)
        self._toolbox = toolbox
        self.execute_in_work = execute_in_work
        if cmd_line_args is None:
            cmd_line_args = []
        self.cmd_line_args = cmd_line_args
        self._cmdline_args_model = ToolCommandLineArgsModel(self)
        self._specification = self._project.get_specification(specification_name)
        if specification_name and not self._specification:
            self._logger.msg_error.emit(
                f"Tool <b>{self.name}</b> should have a Tool specification "
                f"<b>{specification_name}</b> but it was not found"
            )
        self._group_id = group_id
        self._root_directory = deserialize_path(root_dir, self._project.project_dir)
        self._cmdline_args_model.args_updated.connect(self._push_update_cmd_line_args_command)
        self._populate_cmdline_args_model()
        self._input_file_model = FileListModel(header_label="Available resources", draggable=True)
        self._specification_menu = None
        self._options = options if options is not None else {}
        output_dir = deserialize_path(self._options.get("output_directory"), self._project.project_dir)
        self._output_dir = self.default_output_dir if not output_dir else output_dir
        self._kill_completed_processes = kill_completed_processes
        self._log_process_output = log_process_output
        self._resources_from_upstream = []
        self._resources_from_downstream = []
        self._available_resources = []
        self._req_file_paths = set()
        self._input_files_not_found = None

    def set_up(self):
        execute_in_work = self.execute_in_work
        super().set_up()
        self.do_update_execution_mode(execute_in_work)

    def setup_specification_icon(self, spec_icon_path):
        """Adds a specification icon as the child of this item's icon."""
        if not self._specification:
            return
        icon = self.get_icon()
        icon.add_specification_icon(spec_icon_path)

    def update_specification_icon(self):
        """Updates the spec icon when a Tool is selected or when the specification is changed."""
        icon = self.get_icon()
        if icon.spec_item is not None:
            icon.remove_specification_icon()
        spec_icon = ItemInfo.specification_icon(self.specification())
        self.setup_specification_icon(spec_icon)

    @property
    def group_id(self):
        return self._group_id

    @property
    def root_dir(self):
        return self._root_directory

    @property
    def default_output_dir(self):
        return os.path.join(self.data_dir, TOOL_OUTPUT_DIR)

    def _get_options_widget(self):
        """Returns a widget to specify the options for this tool.
        It is embedded in the ui in ``self._update_tool_ui()``.

        Returns:
            OptionsWidget
        """
        # At the moment only Julia has options, but the code is made generic
        constructors = {"julia": JuliaOptionsWidget}  # Add others as needed
        tooltype = self.specification().tooltype
        constructor = constructors.get(tooltype)
        if constructor is None:
            return None
        if tooltype not in self._properties_ui.options_widgets:
            self._properties_ui.options_widgets[tooltype] = constructor()
        options_widget = self._properties_ui.options_widgets[tooltype]
        options_widget.set_tool(self)
        options_widget.do_update_options(self._options)
        return options_widget

    @staticmethod
    def item_type():
        """See base class."""
        return ItemInfo.item_type()

    def make_signal_handler_dict(self):
        """Returns a dictionary of all shared signals and their handlers.
        This is to enable simpler connecting and disconnecting."""
        s = super().make_signal_handler_dict()
        s[self._properties_ui.toolButton_tool_specification.clicked] = self.show_specification_window
        s[self._properties_ui.pushButton_tool_results.clicked] = self._open_results_directory
        s[self._properties_ui.comboBox_tool.textActivated] = self.update_specification
        s[self._properties_ui.radioButton_execute_in_work.toggled] = self.update_execution_mode
        s[self._properties_ui.toolButton_add_file_path_arg.clicked] = self._add_selected_file_path_args
        s[self._properties_ui.toolButton_remove_arg.clicked] = self._remove_arg
        s[self._properties_ui.treeView_input_files.selectionModel().selectionChanged] = (
            self._update_add_args_button_enabled
        )
        s[self._properties_ui.treeView_cmdline_args.selectionModel().selectionChanged] = (
            self._update_remove_args_button_enabled
        )
        s[self._properties_ui.lineEdit_group_id.editingFinished] = self._set_group_id
        s[self._properties_ui.lineEdit_root_directory.editingFinished] = self._set_root_directory
        s[self._properties_ui.toolButton_browse_root_directory.clicked] = self._browse_root_directory
        s[self._properties_ui.lineEdit_result_directory.editingFinished] = self._set_result_directory
        s[self._properties_ui.toolButton_browse_result_directory.clicked] = self._browse_result_directory
        s[self._properties_ui.kill_consoles_check_box.clicked] = self._set_kill_completed_processes
        s[self._properties_ui.log_process_output_check_box.clicked] = self._set_log_process_output
        return s

    @Slot()
    def _set_group_id(self):
        """Pushes a command to update group id whenever the user edits the line edit."""
        group_id = self._properties_ui.lineEdit_group_id.text()
        if not group_id:
            group_id = None
        if self._group_id == group_id:
            return
        self._toolbox.undo_stack.push(UpdateGroupIdCommand(self.name, group_id, self._project))

    def do_set_group_id(self, group_id):
        """Sets group id."""
        self._group_id = group_id
        if self._active:
            self._properties_ui.lineEdit_group_id.setText(group_id)

    @Slot(QItemSelection, QItemSelection)
    def _update_add_args_button_enabled(self, _selected, _deselected):
        self._do_update_add_args_button_enabled()

    def _do_update_add_args_button_enabled(self):
        enabled = self._properties_ui.treeView_input_files.selectionModel().hasSelection()
        self._properties_ui.toolButton_add_file_path_arg.setEnabled(enabled)

    @Slot(QItemSelection, QItemSelection)
    def _update_remove_args_button_enabled(self, _selected, _deselected):
        self._do_update_remove_args_button_enabled()

    def _do_update_remove_args_button_enabled(self):
        enabled = self._properties_ui.treeView_cmdline_args.selectionModel().hasSelection()
        self._properties_ui.toolButton_remove_arg.setEnabled(enabled)

    def _update_kill_consoles_check_box_enabled(self):
        """Updates the enabled state of 'Kill consoles after execution' check box."""
        enabled = self._specification is None or self._specification.tooltype in ("python", "julia")
        self._properties_ui.kill_consoles_check_box.setEnabled(enabled)

    def restore_selections(self):
        """Restore selections into shared widgets when this project item is selected."""
        self._properties_ui.treeView_input_files.setModel(self._input_file_model)
        self._properties_ui.treeView_cmdline_args.setModel(self._cmdline_args_model)
        self._properties_ui.treeView_cmdline_args.expandAll()
        self.update_execute_in_work_button()
        self._properties_ui.label_jupyter.elided_mode = Qt.TextElideMode.ElideMiddle
        self._properties_ui.label_jupyter.hide()
        self._update_tool_ui()
        self._do_update_add_args_button_enabled()
        self._do_update_remove_args_button_enabled()
        self._properties_ui.lineEdit_group_id.setText(self._group_id)
        self._properties_ui.lineEdit_root_directory.setText(self._root_directory)
        output_dir_displayed = self._output_dir.replace("/", os.sep)
        if same_path(self._output_dir, self.default_output_dir):
            self._properties_ui.lineEdit_result_directory.clear()
        else:  # Set text only if results directory is not default
            self._properties_ui.lineEdit_result_directory.setText(output_dir_displayed)
        self._properties_ui.lineEdit_result_directory.setPlaceholderText(output_dir_displayed)
        self._properties_ui.lineEdit_result_directory.setToolTip(output_dir_displayed)

    @Slot(bool)
    def _browse_root_directory(self, _=False):
        """Calls static method that shows a file browser for selecting a Tool spec main program file root directory."""
        select_root_directory(self._toolbox, self._properties_ui.lineEdit_root_directory, self._project.project_dir)
        self._set_root_directory()

    @Slot()
    def _set_root_directory(self):
        """Pushes a command to update root directory whenever the user edits the line edit."""
        root_dir = self._properties_ui.lineEdit_root_directory.text()
        if not root_dir:
            root_dir = None
        if self._root_directory == root_dir:
            return
        self._toolbox.undo_stack.push(UpdateRootDirCommand(self.name, root_dir, self._project))

    def do_set_root_directory(self, root_dir):
        """Sets root directory."""
        self._root_directory = root_dir
        if self._active:
            self._properties_ui.lineEdit_root_directory.setText(root_dir)

    @Slot(bool)
    def _browse_result_directory(self, _=False):
        """Calls static method that shows a file browser for selecting a result directory."""
        select_root_directory(self._toolbox, self._properties_ui.lineEdit_result_directory, self._project.project_dir)
        self._set_result_directory()

    @Slot()
    def _set_result_directory(self):
        """Pushes a command to update result directory whenever the user edits the line edit."""
        result_dir = self._properties_ui.lineEdit_result_directory.text().strip()
        if same_path(self._output_dir, result_dir):  # When user picks the same folder again
            self._properties_ui.lineEdit_result_directory.clear()
            return
        if not result_dir and same_path(self._output_dir, self.default_output_dir):
            return
        self._toolbox.undo_stack.push(UpdateResultDirCommand(self.name, result_dir, self._project))

    def do_set_result_directory(self, result_dir):
        """Sets results directory."""
        self._output_dir = self.default_output_dir if not result_dir else result_dir
        self._options["output_directory"] = serialize_path(self._output_dir, self._project.project_dir)
        if self._active:
            output_dir_displayed = self._output_dir.replace("/", os.sep)
            if not result_dir:
                self._properties_ui.lineEdit_result_directory.clear()
                self._properties_ui.lineEdit_result_directory.setPlaceholderText(output_dir_displayed)
            else:
                self._properties_ui.lineEdit_result_directory.setText(output_dir_displayed)
                self._properties_ui.lineEdit_result_directory.setToolTip(output_dir_displayed)
            self._properties_ui.lineEdit_result_directory.setToolTip(output_dir_displayed)

    @Slot(bool)
    def show_specification_window(self, _=True):
        """Opens the settings window."""
        self._toolbox.show_specification_form(self.item_type(), self.specification(), self)

    @Slot(bool)
    def update_execution_mode(self, checked):
        """Pushes a new UpdateToolExecuteInWorkCommand to the toolbox stack."""
        self._toolbox.undo_stack.push(UpdateToolExecuteInWorkCommand(self.name, checked, self._project))

    def do_update_execution_mode(self, execute_in_work):
        """Updates execute_in_work setting."""
        if self.execute_in_work == execute_in_work:
            return
        self.execute_in_work = execute_in_work
        self.update_execute_in_work_button()

    def update_execute_in_work_button(self):
        """Sets execute in work radio button check state according to
        execute_in_work instance variable."""
        if not self._active:
            return
        self._properties_ui.radioButton_execute_in_work.blockSignals(True)
        self._properties_ui.radioButton_execute_in_source.blockSignals(True)
        if self.execute_in_work:
            self._properties_ui.radioButton_execute_in_work.setChecked(True)
        else:
            self._properties_ui.radioButton_execute_in_source.setChecked(True)
        self._properties_ui.radioButton_execute_in_work.blockSignals(False)
        self._properties_ui.radioButton_execute_in_source.blockSignals(False)

    @Slot(str)
    def update_specification(self, text):
        """Update Tool specification according to selection in the specification comboBox.

        Args:
            text (str): Tool specification name in the comboBox
        """
        spec = self._project.get_specification(text)
        self.set_specification(spec)

    @Slot(bool)
    def _remove_arg(self, _=False):
        removed_rows = [index.row() for index in self._properties_ui.treeView_cmdline_args.selectedIndexes()]
        cmd_line_args = [arg for row, arg in enumerate(self.cmd_line_args) if row not in removed_rows]
        self._push_update_cmd_line_args_command(cmd_line_args)

    @Slot(bool)
    def _add_selected_file_path_args(self, _=False):
        new_args = [LabelArg(index.data()) for index in self._properties_ui.treeView_input_files.selectedIndexes()]
        self._push_update_cmd_line_args_command(self.cmd_line_args + new_args)

    @Slot(list)
    def _push_update_cmd_line_args_command(self, cmd_line_args):
        if self.cmd_line_args == cmd_line_args:
            return
        self._toolbox.undo_stack.push(UpdateCmdLineArgsCommand(self.name, cmd_line_args, self._project))

    def update_cmd_line_args(self, cmd_line_args):
        """Updates instance cmd line args list and sets the list as text to the line edit widget.

        Args:
            cmd_line_args (list): Tool cmd line args
        """
        self.cmd_line_args = cmd_line_args
        self._populate_cmdline_args_model()
        self._check_notifications()

    def _populate_cmdline_args_model(self):
        spec_args = [CmdLineArg(arg) for arg in self._specification.cmdline_args] if self._specification else []
        tool_args = self.cmd_line_args
        self._cmdline_args_model.reset_model(spec_args, tool_args)
        if self._active:
            self._properties_ui.treeView_cmdline_args.setFocus()

    def undo_specification(self):
        if self._specification is None:
            return None
        undo_spec = self._specification.clone()
        return undo_spec

    def do_set_specification(self, specification):
        """see base class"""
        if not super().do_set_specification(specification):
            return False
        self._populate_cmdline_args_model()
        if self._active:
            self._update_tool_ui()
        self._resources_to_successors_changed()
        self._check_notifications()
        self.update_specification_icon()
        return True

    def update_options(self, options):
        """Pushes a new UpdateToolOptionsCommand to the toolbox undo stack."""
        self._toolbox.undo_stack.push(UpdateToolOptionsCommand(self.name, options, self._options, self._project))

    def do_set_options(self, options):
        """Sets options for this tool.

        Args:
            options (dict): The new options dictionary, must include *ALL* the options, not only changed ones.
        """
        self._options = options
        if self._active:
            _ = self._get_options_widget()

    @Slot(bool)
    def _set_kill_completed_processes(self, kill_completed_processes):
        """Pushes a UpdateKillCompletedProcesses to the undo stack

        Args:
            kill_completed_processes (bool): whether to kill completed persistent processes after execution
        """
        self._toolbox.undo_stack.push(UpdateKillCompletedProcesses(self.name, kill_completed_processes, self._project))

    def do_update_kill_completed_processes(self, kill_completed_processes):
        """Updates the kill_completed_processes flag.

        Args:
            kill_completed_processes (bool): whether to kill completed persistent processes after execution
        """
        self._kill_completed_processes = kill_completed_processes
        if self._active:
            self._properties_ui.kill_consoles_check_box.setChecked(kill_completed_processes)

    @Slot(bool)
    def _set_log_process_output(self, log_process_output):
        """Pushes a UpdateLogProcessOutput to the undo stack

        Args:
            log_process_output (bool): whether to kill completed persistent processes after execution
        """
        self._toolbox.undo_stack.push(UpdateLogProcessOutput(self.name, log_process_output, self._project))

    def do_update_log_process_output(self, log_process_output):
        """Updates the log_process_output flag.

        Args:
            log_process_output (bool): whether to kill completed persistent processes after execution
        """
        self._log_process_output = log_process_output
        if self._active:
            self._properties_ui.log_process_output_check_box.setChecked(log_process_output)

    def _update_tool_ui(self):
        """Updates Tool properties UI. Used when Tool specification is changed."""
        options_widget = self._properties_ui.horizontalLayout_options.takeAt(0)
        if options_widget:
            options_widget.widget().hide()
        if not self.specification():
            self._properties_ui.comboBox_tool.setCurrentIndex(-1)
            self.do_update_execution_mode(True)
            self._properties_ui.toolButton_tool_specification.setMenu(None)
            return
        self._properties_ui.comboBox_tool.setCurrentText(self.specification().name)
        self._update_specification_menu()
        self._properties_ui.toolButton_tool_specification.setMenu(self._specification_menu)
        options_widget = self._get_options_widget()
        if options_widget:
            self._properties_ui.horizontalLayout_options.addWidget(options_widget)
            options_widget.show()
        self._show_execution_settings()
        self._properties_ui.kill_consoles_check_box.setChecked(self._kill_completed_processes)
        self._update_kill_consoles_check_box_enabled()
        self._properties_ui.log_process_output_check_box.setChecked(self._log_process_output)

    def _show_execution_settings(self):
        """Updates the label in Tool properties to show the selected execution settings for this Tool."""
        tstype = self._specification.tooltype
        if tstype in {"python", "julia"}:
            self.specification().init_execution_settings()
            k_spec_name = self.specification().execution_settings["kernel_spec_name"]
            env = self.specification().execution_settings["env"]
            use_console = self.specification().execution_settings["use_jupyter_console"]
            self._properties_ui.label_jupyter.show()
            if not use_console:
                exe = self.specification().execution_settings["executable"]
                if tstype == "python":
                    path = exe if exe else resolve_python_interpreter(self._project.app_settings)
                else:
                    path = exe if exe else resolve_julia_executable(self._project.app_settings)
                self._properties_ui.label_jupyter.setText(f"[Basic console] {path}")
            else:
                env = "" if not env else f"[{env}]"
                self._properties_ui.label_jupyter.setText(f"[Jupyter Console] {k_spec_name} {env}")

    def _update_specification_menu(self):
        spec_model_index = self._toolbox.specification_model.specification_index(self.specification().name)
        self._specification_menu = ToolSpecificationMenu(self._toolbox, spec_model_index, self)
        self._specification_menu.setTitle("Specification...")

    @Slot(bool)
    def _open_results_directory(self, _):
        """Open output directory in file browser."""
        if not os.path.exists(self._output_dir):
            self._logger.msg_warning.emit(f"Tool <b>{self.name}</b> has no results. Click Execute to generate them.")
            return
        url = "file:///" + self._output_dir
        # noinspection PyTypeChecker, PyCallByClass, PyArgumentList
        res = open_url(url)
        if not res:
            self._logger.msg_error.emit(f"Failed to open directory: {self._output_dir}")

    def specification(self):
        """Returns Tool specification."""
        return self._specification

    def resources_for_direct_successors(self):
        """See base class"""
        return scan_for_resources(self, self.specification(), self._output_dir)

    def _find_input_files(self, resources):
        """Iterates files in required input files model and looks for them in the given resources.

        Args:
            resources (list): resources available

        Returns:
            Dictionary mapping required files to path where they are found, or to None if not found
        """
        if not self.specification():
            return {}
        file_paths = {}
        for req_file_path in self.specification().inputfiles:
            # Just get the filename if there is a path attached to the file
            _, filename = os.path.split(req_file_path)
            if not filename:
                # It's a directory
                continue
            file_paths[req_file_path] = find_file(filename, resources, one_file=True)
        return file_paths

    def _check_notifications(self):
        """See base class."""
        self.clear_notifications()
        self._check_write_index()
        if not self.specification():
            self.add_notification("This Tool does not have a specification. Set it in the Tool Properties Panel.")
        elif self.specification().includes and not self.specification().path:
            n = self.specification().name
            self.add_notification(
                f"Tool specification <b>{n}</b> path does not exist. Fix this in Tool specification editor."
            )
        resources = self._resources_from_upstream + self._resources_from_downstream
        resources_changed = resources != self._available_resources
        req_files_changed = False
        if self.specification():
            required_files = self.specification().inputfiles
            req_files_changed = required_files != self._req_file_paths
            # No need to perform slow check for if all the file requirements are/aren't still satisfied
            # if no changes have been made to the Tool's requirements or available resources
            if required_files and (req_files_changed or resources_changed):
                file_paths = self._find_input_files(resources)
                file_paths = flatten_file_path_duplicates(file_paths, self._logger)
                self._input_files_not_found = [k for k, v in file_paths.items() if v is None]
            # Check that main program file exists. If not, log a message with an anchor to find it
            if len(self.specification().includes) > 0:
                filename = self.specification().includes[0]
                full_path = os.path.join(self.specification().path, filename)
                if not os.path.isfile(full_path):
                    self.add_notification(
                        f"Tool spec <b>{self.specification().name}</b> won't work because "
                        f"main program file <b>{full_path}</b> doesn't exist."
                    )
        if self._input_files_not_found:
            self.add_notification(
                f"File(s) {', '.join(self._input_files_not_found)} needed to execute this Tool are not provided"
                f"by any input item. Connect items that provide the required files to this Tool."
            )
        if resources_changed or req_files_changed:
            # Only check for duplicate files in available resources when the resources have changed
            if resources_changed:
                self._available_resources = resources
                duplicates = self._input_file_model.duplicate_paths()
                if duplicates:
                    self.add_notification(f"Duplicate input files:<br>{'<br>'.join(duplicates)}")
            if req_files_changed:
                self._req_file_paths = required_files
        missing_args = ", ".join(arg.arg for arg in self.cmd_line_args if isinstance(arg, LabelArg) and arg.missing)
        if missing_args:
            self.add_notification(
                f"The following command line argument(s) don't match any available resources: {missing_args}"
            )

    def handle_execution_successful(self, execution_direction):
        """See base class."""
        super().handle_execution_successful(execution_direction)
        if execution_direction != ExecutionDirection.FORWARD:
            return
        self._resources_to_successors_changed()

    def upstream_resources_updated(self, resources):
        """See base class."""
        self._resources_from_upstream = resources
        self._update_files_and_cmd_line_args()

    def replace_resources_from_upstream(self, old, new):
        """See base class."""
        self._replace_resources(old, new, self._resources_from_upstream)

    def downstream_resources_updated(self, resources):
        """See base class."""
        self._resources_from_downstream = resources
        self._update_files_and_cmd_line_args()

    def replace_resources_from_downstream(self, old, new):
        """See base class."""
        self._replace_resources(old, new, self._resources_from_downstream)

    def _replace_resources(self, old_resources, new_resources, resource_list):
        """Replaces resources.

        Modifies ``resource_list`` in-place!

        Args:
            old_resources (list of ProjectItemResource): old resources
            new_resources (list of ProjectItemResource): new resources
            resource_list (list of ProjectItemResource): current downstream or upstream resources
        """
        for old, new in zip(old_resources, new_resources):
            for i, resource in enumerate(resource_list):
                if resource == old:
                    resource_list[i] = new
                    break
            for i, arg in enumerate(self.cmd_line_args):
                if arg.arg == old.label:
                    self.cmd_line_args[i] = LabelArg(new.label)
        self._input_file_model.update(self._resources_from_upstream + self._resources_from_downstream)
        self.update_cmd_line_args(self.cmd_line_args)

    def _update_files_and_cmd_line_args(self):
        """Updates the file model and command line arguments."""
        resources = self._resources_from_upstream + self._resources_from_downstream
        self._input_file_model.update(resources)
        update_args = []
        resource_labels = {resource.label for resource in resources}
        for arg in self.cmd_line_args:
            if arg.arg in resource_labels:
                arg.missing = False
            elif isinstance(arg, LabelArg):
                # Arg may refer to legacy Data connection resource labels that weren't prefixed by <project> or <data>
                legacy_arg_converted = False
                for label in resource_labels:
                    if label.startswith("<") and label.endswith(arg.arg):
                        legacy_arg_converted = True
                        arg.arg = label
                        break
                if not legacy_arg_converted:
                    arg.missing = True
            update_args.append(arg)
        self.update_cmd_line_args(update_args)

    def item_dict(self):
        """Returns a dictionary corresponding to this item."""
        d = super().item_dict()
        if not self.specification():
            d["specification"] = ""
        else:
            d["specification"] = self.specification().name
        d["execute_in_work"] = self.execute_in_work
        d["cmd_line_args"] = [arg.to_dict() for arg in self.cmd_line_args]
        if not same_path(self._output_dir, self.default_output_dir):
            self._options["output_directory"] = serialize_path(self._output_dir, self._project.project_dir)
        else:
            self._options.pop("output_directory", None)
        if self._options:
            d["options"] = self._options
        d["kill_completed_processes"] = self._kill_completed_processes
        d["log_process_output"] = self._log_process_output
        if self._group_id:
            d["group_id"] = self._group_id
        if self._root_directory:
            d["root_directory"] = serialize_path(self._root_directory, self._project.project_dir)
        return d

    @staticmethod
    def item_dict_local_entries():
        """See base class."""
        return [("root_directory",), ("options",)]

    @staticmethod
    def from_dict(name, item_dict, toolbox, project):
        """See base class."""
        description, x, y = DBWriterItemBase.parse_item_dict(item_dict)
        specification_name = item_dict.get("specification", "")
        execute_in_work = item_dict.get("execute_in_work", True)
        cmd_line_args = item_dict.get("cmd_line_args", [])
        cmd_line_args = [make_cmd_line_arg(arg) for arg in cmd_line_args]
        options = item_dict.get("options", {})
        kill_completed_processes = item_dict.get("kill_completed_processes", False)
        log_process_output = item_dict.get("log_process_output", False)
        group_id = item_dict.get("group_id")
        root_dir = item_dict.get("root_directory", "")
        return Tool(
            name,
            description,
            x,
            y,
            toolbox,
            project,
            specification_name,
            execute_in_work,
            cmd_line_args,
            options,
            kill_completed_processes,
            log_process_output,
            group_id,
            root_dir,
        )

    def rename(self, new_name, rename_data_dir_message):
        """See base class."""
        if not super().rename(new_name, rename_data_dir_message):
            return False
        output_dir = deserialize_path(self._options.get("output_directory"), self._project.project_dir)
        self._output_dir = self.default_output_dir if not output_dir else output_dir
        return True

    def notify_destination(self, source_item):
        """See base class."""
        if source_item.item_type() == "Data Store":
            self._logger.msg.emit(
                f"Link established. Data Store <b>{source_item.name}</b> url will "
                f"be passed to Tool <b>{self.name}</b> when executing."
            )
        elif source_item.item_type() == "Data Connection":
            self._logger.msg.emit(
                f"Link established. Tool <b>{self.name}</b> will look for input "
                f"files from <b>{source_item.name}</b>'s references and data directory."
            )
        elif source_item.item_type() == "Exporter":
            self._logger.msg.emit(
                f"Link established. The file exported by <b>{source_item.name}</b> are now "
                f"available in <b>{self.name}</b>."
            )
        elif source_item.item_type() in ["Data Transformer", "Tool"]:
            self._logger.msg.emit("Link established")
        else:
            super().notify_destination(source_item)

    def actions(self):
        # pylint: disable=attribute-defined-outside-init
        if self.specification() is not None:
            if self._specification_menu is None:
                self._update_specification_menu()
            self._actions = [self._specification_menu.menuAction()]
        else:
            action = QAction("New specification")
            action.triggered.connect(self.show_specification_window)
            self._actions = [action]
        self._actions.append(QAction("Open results directory"))
        self._actions[-1].triggered.connect(self._open_results_directory)
        return self._actions

    @staticmethod
    def upgrade_v1_to_v2(item_name, item_dict):
        """Upgrades item's dictionary from v1 to v2.

        Changes:
        - 'tool' key is renamed to 'specification'

        Args:
            item_name (str): item's name
            item_dict (dict): Version 1 item dictionary

        Returns:
            dict: Version 2 Tool dictionary
        """
        item_dict["specification"] = item_dict.pop("tool", "")
        return item_dict



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains unit tests for Exporter's ``utils`` module."""
import os.path
import unittest
from spine_engine.project_item.project_item_resource import url_resource
from spine_items.exporter.output_channel import OutputChannel
from spine_items.exporter.utils import output_database_resources


class TestOutputDatabaseResources(unittest.TestCase):
    def test_creates_url_resources_for_channels_with_urls(self):
        item_name = "test exporter"
        channels = [
            OutputChannel("label 1", item_name, "file.csv"),
            OutputChannel(
                "label 1", item_name, "out database", {"dialect": "sqlite", "database": "/path/to/db.sqlite"}
            ),
        ]
        resources = output_database_resources(item_name, channels)
        expected_url = "sqlite:///" + os.path.normcase(os.path.abspath("/path/to/db.sqlite"))
        self.assertEqual(resources, [url_resource(item_name, expected_url, "out database")])


if __name__ == "__main__":
    unittest.main()



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Toolbox.
# Spine Toolbox is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

""" Contains ImportPreviewWindow class. """
import fnmatch
import json
from operator import attrgetter
import os
from PySide6.QtCore import QItemSelectionModel, QModelIndex, Qt, QTimer, Signal, Slot
from PySide6.QtWidgets import QDialog, QDialogButtonBox, QFileDialog, QListWidget, QMessageBox, QVBoxLayout
from spinedb_api.exception import ReaderError
from spinedb_api.helpers import remove_credentials_from_url
from spinedb_api.spine_io.gdx_utils import find_gams_directory
from spinedb_api.spine_io.importers.reader import Reader
from spinedb_api.spine_io.importers.sqlalchemy_reader import SQLAlchemyReader
from spinetoolbox.config import APPLICATION_PATH
from spinetoolbox.helpers import get_open_file_name_in_last_dir
from spinetoolbox.project_item.specification_editor_window import SpecificationEditorWindowBase
from ...utils import convert_to_sqlalchemy_url
from ...widgets import UrlSelectorDialog
from ..commands import RestoreMappingsFromDict
from ..connection_manager import ConnectionManager
from ..executable_item import READER_NAME_TO_CLASS
from ..importer_specification import ImporterSpecification
from ..mvcmodels.mappings_model import MappingsModel
from ..mvcmodels.source_list_selection_model import SourceListSelectionModel
from .import_mapping_options import ImportMappingOptions
from .import_mappings import ImportMappings
from .import_sources import ImportSources


class _ReaderProblemInMapping(Exception):
    """Raised when mapping has no reader or the reader looks different to file type."""

    def __init__(self, reader_in_mapping):
        """
        Args:
            reader_in_mapping (type, optional): reader class defined in mapping
        """
        self.reader_in_mapping = reader_in_mapping


class ImportEditorWindow(SpecificationEditorWindowBase):
    """A QMainWindow to let users define Mappings for an Importer item."""

    connection_failed = Signal(str)

    _FILE_LESS = "anonymous"
    """Name of the 'file-less' input."""

    class _FileLessReader(Reader):
        """A connector that has no tables or contents, used for the file-less mode."""

        DISPLAY_NAME = "<unspecified>"
        FILE_EXTENSIONS = ""
        OPTIONS = {}

        def connect_to_source(self, source, **extras):
            pass

        def disconnect(self):
            pass

        def get_tables_and_properties(self):
            return {}

        def get_data_iterator(self, table, options, max_rows=-1):
            return iter([]), ()

    def __init__(self, toolbox, specification, item=None, input=None, input_extras=None):
        """
        Args:
            toolbox (QMainWindow): ToolboxUI class
            specification (ImporterSpecification, optional): Importer specification
            item (Importer, optional): Linked Importer item
            input (str, optional): Importee file path or URL; if None, work in file-less mode
            input_extras (dict, optional): Additional input settings such as database schema
        """
        super().__init__(toolbox, specification, item)
        self._input = input if input is not None else self._FILE_LESS
        self._input_extras = input_extras if input_extras is not None else {}
        self._mappings_model = MappingsModel(self._undo_stack, self)
        self._mappings_model.rowsInserted.connect(self._reselect_source_table)
        self._ui.source_list.setModel(self._mappings_model)
        self._ui.source_list.setSelectionModel(SourceListSelectionModel(self._mappings_model))
        self._ui.mapping_list.setModel(self._mappings_model)
        self._ui.mapping_list.setRootIndex(self._mappings_model.dummy_parent())
        self._ui.mapping_spec_table.setModel(self._mappings_model)
        self._ui.mapping_spec_table.setRootIndex(self._mappings_model.dummy_parent())
        self._connection_manager = None
        self._memoized_reader = None
        self._import_mappings = ImportMappings(self._mappings_model, self._ui, self._undo_stack, self)
        self._import_mapping_options = ImportMappingOptions(self._mappings_model, self._ui, self._undo_stack)
        self._import_sources = ImportSources(self._mappings_model, self._ui, self._undo_stack, self)
        self._set_input_text()
        self._ui.input_path_line_edit.editingFinished.connect(self._read_input_path_from_line)
        self._ui.input_path_line_edit.textEdited.connect(self._maybe_switch_to_file_less_mode)
        self._ui.browse_inputs_button.clicked.connect(self._show_open_file_dialog)
        self._ui.import_mappings_action.triggered.connect(self.import_mapping_from_file)
        self._ui.export_mappings_action.triggered.connect(self.export_mapping_to_file)
        self._ui.switch_input_type_action.triggered.connect(self._switch_input_type)
        self.connection_failed.connect(self.show_error)
        self._import_sources.preview_data_updated.connect(self._import_mapping_options.set_num_available_columns)
        self._mappings_model.restore(self.specification.mapping if self.specification is not None else {})
        QTimer.singleShot(0, self.start_ui)

    def is_file_less(self):
        return not self._ui.input_path_line_edit.text()

    @property
    def settings_group(self):
        return "mappingPreviewWindow"

    def _save(self, exiting=None):
        """See base class."""
        if super()._save(exiting):
            self._import_mappings.specification_saved()
            return True

    @property
    def _duplicate_kwargs(self):
        return {"input": self._input}

    def _make_ui(self):
        from ..ui.import_editor_window import Ui_MainWindow  # pylint: disable=import-outside-toplevel

        return Ui_MainWindow()

    def _make_new_specification(self, spec_name):
        mappings_dict = self._mappings_model.store()
        mappings_dict.update(self._import_sources.store_connectors())
        description = self._spec_toolbar.description()
        return ImporterSpecification(spec_name, mappings_dict, description)

    def _populate_main_menu(self):
        super()._populate_main_menu()
        menu = self._spec_toolbar.menu
        before = self._spec_toolbar.save_action
        menu.insertAction(before, self._ui.switch_input_type_action)
        menu.insertSeparator(before)
        menu.insertActions(before, [self._ui.import_mappings_action, self._ui.export_mappings_action])
        menu.insertSeparator(before)

    def _set_input_text(self):
        """Sets input path/URL to the input line edit cleaning credentials from URLs."""
        if self._input == self._FILE_LESS:
            self._ui.input_path_line_edit.clear()
            return
        label = remove_credentials_from_url(self._input) if self._is_url(self._input) else self._input
        self._ui.input_path_line_edit.setText(label)

    @Slot(bool)
    def _show_open_file_dialog(self, _=False):
        if self._is_database_reader(self._connection_manager.connection):
            url = self._get_input_url()
            if url is None:
                return
            self._input = str(convert_to_sqlalchemy_url(url, logger=self))
            schema = url["schema"]
            self._input_extras = {"schema": schema if schema else None}
        else:
            file_path = self._get_input_file_path()
            if not file_path:
                return
            self._input = file_path
            self._input_extras = None
        self._set_input_text()
        self.start_ui()

    def _get_input_url(self):
        selector = UrlSelectorDialog(self._toolbox.qsettings(), False, self._toolbox, parent=self)
        selector.exec()
        if selector.result() != QDialog.DialogCode.Accepted:
            return None
        return selector.url_dict()

    def _get_input_file_path(self):
        filter_ = ";;".join(["*.*"] + [conn.FILE_EXTENSIONS for conn in READER_NAME_TO_CLASS.values()])
        key = f"selectInputDataFileFor{self.specification.name if self.specification else None}"
        filepath, _ = get_open_file_name_in_last_dir(
            self._toolbox.qsettings(),
            key,
            self,
            "Select an input data file to define the specification",
            APPLICATION_PATH,
            filter_=filter_,
        )
        return filepath

    @staticmethod
    def _is_database_reader(reader):
        """Tests if reader class works with database URLs.

        Args:
            reader (type): reader class to test
        """
        return reader.__name__ == SQLAlchemyReader.__name__

    @Slot(bool)
    def _switch_input_type(self, _=False):
        if self.specification:
            self.specification.mapping.pop("source_type", None)
        self._memoized_reader = None
        self.start_ui()

    def _get_reader_from_mapping(self, input_path):
        """Reads reader for given input from mapping.

        Args:
            input_path (str): importee file path or URL

        Returns:
            type: reader class
        """
        if not self.specification:
            raise _ReaderProblemInMapping(None)
        mapping = self.specification.mapping
        source_type = mapping.get("source_type")
        if source_type is None:
            raise _ReaderProblemInMapping(None)
        reader = READER_NAME_TO_CLASS[source_type]
        file_extensions = reader.FILE_EXTENSIONS.split(";;")
        if input_path != self._FILE_LESS and not any(fnmatch.fnmatch(input_path, ext) for ext in file_extensions):
            if reader is SQLAlchemyReader and self._is_url(input_path):
                return reader
            raise _ReaderProblemInMapping(reader)
        return reader

    @Slot()
    def _read_input_path_from_line(self):
        """Sets input from input path line edit."""
        label = self._ui.input_path_line_edit.text()
        if self._input == self._FILE_LESS:
            if not label:
                return
            self._input = label
        else:
            if self._is_url(self._input):
                if label == remove_credentials_from_url(self._input):
                    return
                self._input_extras = None
            elif label == self._input:
                return
            self._input = label if label else self._FILE_LESS
        self.start_ui()

    @Slot(str)
    def _maybe_switch_to_file_less_mode(self, text):
        """Switches to file-less mode if text has been cleared.

        Args:
            text (str): text
        """
        if text or self._input == self._FILE_LESS:
            return
        self._input = self._FILE_LESS
        self._input_extras = None
        self.start_ui()

    def start_ui(self):
        """Connects to input and fills the tables and lists with data."""
        try:
            reader = self._get_reader_from_mapping(self._input)
        except _ReaderProblemInMapping as reader_problem:
            if reader_problem.reader_in_mapping is not None:
                QMessageBox.warning(
                    self,
                    "Verify input type",
                    f"Input type is set to {reader_problem.reader_in_mapping.DISPLAY_NAME} but the input looks incompatible. "
                    "You will be prompted to verify the type.",
                )
            reader = self._get_reader(self._input)
            if not reader:
                return
            is_db_reader = self._is_database_reader(reader)
            is_url_input = self._is_url(self._input)
            if (is_db_reader and not is_url_input) or (not is_db_reader and is_url_input):
                self._ui.input_path_line_edit.clear()
                self._input = self._FILE_LESS
                self._input_extras = None
            if reader_problem.reader_in_mapping is not None and reader is not reader_problem.reader_in_mapping:
                QMessageBox.information(
                    self,
                    "Input type changed",
                    f"Input type changed from {reader_problem.reader_in_mapping.DISPLAY_NAME} to {reader.DISPLAY_NAME}. "
                    "Don't forget to save the changes.",
                )
                self._undo_stack.resetClean()
        if self._input == self._FILE_LESS:
            self._FileLessReader.__name__ = reader.__name__
            self._FileLessReader.FILE_EXTENSIONS = reader.FILE_EXTENSIONS
            self._FileLessReader.OPTIONS = reader.OPTIONS
            reader = self._FileLessReader
            self._mappings_model.set_tables_editable(True)
        else:
            self._mappings_model.set_tables_editable(False)
        self._ui.switch_input_type_action.setEnabled(True)
        reader_settings = {"gams_directory": _gams_system_directory(self._toolbox)}
        if self._connection_manager:
            self._connection_manager.close_connection()
        self._connection_manager = ConnectionManager(reader, reader_settings, self)
        self._connection_manager.connection_failed.connect(self.connection_failed.emit)
        self._connection_manager.error.connect(self.show_error)
        for header in (self._ui.source_data_table.horizontalHeader(), self._ui.source_data_table.verticalHeader()):
            self._ui.source_list.selectionModel().currentChanged.connect(
                header.set_source_table, Qt.ConnectionType.UniqueConnection
            )
        self._connection_manager.connection_ready.connect(self._handle_connection_ready)
        mapping = self.specification.mapping if self.specification else {}
        self._import_sources.set_connector(self._connection_manager, mapping)
        self._display_input_type(reader.DISPLAY_NAME)
        extras = self._input_extras if self._input_extras is not None else {}
        try:
            self._connection_manager.init_connection(self._input, **extras)
        except ReaderError as error:
            QMessageBox.warning(self, "Failed to read source", f"Couldn't read source: {error}")

    def _display_input_type(self, input_type):
        """Shows connector's name on the ui.

        Args:
            input_type (str): readers's display name
        """
        self._ui.input_type_line_edit.setText(input_type)

    @Slot()
    def _handle_connection_ready(self):
        self._ui.export_mappings_action.setEnabled(True)
        self._ui.import_mappings_action.setEnabled(True)

    def _get_reader(self, input_path):
        """Shows a QDialog to select a reader for the given data input.

        Args:
            input_path (str): Path of the file acting as an importee

        Returns:
            Asynchronous data reader class for the given importee
        """
        if self._memoized_reader:
            return self._memoized_reader
        reader_list = sorted(list(READER_NAME_TO_CLASS.values()), key=attrgetter("DISPLAY_NAME"))
        reader_names = [c.DISPLAY_NAME for c in reader_list]
        dialog = QDialog(self)
        dialog.setLayout(QVBoxLayout())
        reader_list_wg = QListWidget()
        reader_list_wg.addItems(reader_names)
        # Set current item in `reader_list_wg` based on file extension
        row = None
        for k, reader in enumerate(reader_list):
            file_extensions = reader.FILE_EXTENSIONS.split(";;")
            if any(fnmatch.fnmatch(input_path, ext) for ext in file_extensions):
                row = k
                break
        else:
            if self._is_url(input_path):
                row = reader_names.index(SQLAlchemyReader.DISPLAY_NAME)
        if row is not None:
            reader_list_wg.setCurrentRow(row)
        button_box = QDialogButtonBox(QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel)
        button_box.button(QDialogButtonBox.StandardButton.Ok).clicked.connect(dialog.accept)
        button_box.button(QDialogButtonBox.StandardButton.Cancel).clicked.connect(dialog.reject)
        reader_list_wg.doubleClicked.connect(dialog.accept)
        dialog.layout().addWidget(reader_list_wg)
        dialog.layout().addWidget(button_box)
        spec_name = self._spec_toolbar.name()
        if not spec_name:
            spec_name = "unnamed specification"
        dialog.setWindowTitle(f"Select input type for {spec_name}")
        answer = dialog.exec()
        if answer != QDialog.DialogCode.Accepted:
            return None
        row = reader_list_wg.currentIndex().row()
        if row < 0:
            return None
        connector = self._memoized_reader = reader_list[row]
        return connector

    @Slot()
    def import_mapping_from_file(self):
        """Imports mapping spec from a user selected .json file to the preview window."""
        start_dir = self._toolbox.project().project_dir
        # noinspection PyCallByClass
        filename = QFileDialog.getOpenFileName(
            self, "Import mapping specification", start_dir, "Import mapping (*.json)"
        )
        if not filename[0]:
            return
        with open(filename[0]) as file_p:
            try:
                settings = json.load(file_p)
            except json.JSONDecodeError:
                self._show_status_bar_msg(f"Could not open {filename[0]}")
                return
        expected_options = ("table_mappings", "table_types", "table_row_types", "table_options", "selected_tables")
        if not isinstance(settings, dict) or not any(key in expected_options for key in settings.keys()):
            self._show_status_bar_msg(f"{filename[0]} does not contain and import mapping")
        self._undo_stack.push(RestoreMappingsFromDict(self._import_sources, self._mappings_model, settings))
        self._show_status_bar_msg(f"Mapping loaded from {filename[0]}")

    @Slot()
    def export_mapping_to_file(self):
        """Exports all mapping specs in current preview window to .json file."""
        start_dir = self._toolbox.project().project_dir
        # noinspection PyCallByClass
        filename = QFileDialog.getSaveFileName(
            self, "Export mapping spec to a file", start_dir, "Import mapping (*.json)"
        )
        if not filename[0]:
            return
        with open(filename[0], "w") as file_p:
            mappings_dict = self._mappings_model.store()
            mappings_dict.update(self._import_sources.store_connectors())
            json.dump(mappings_dict, file_p)
        self._show_status_bar_msg(f"Mapping saved to: {filename[0]}")

    @Slot(QModelIndex, int, int)
    def _reselect_source_table(self, parent, first, last):
        """Selects added source table.

        This is a workaround to get the correct source table selected after a new one has been added
        since the source table view doesn't seem to update the current index correctly in this case.

        Args:
            parent (QModelIndex): parent index
            first (int): first new row
            last (int): last new row
        """
        if parent.isValid():
            return
        index = self._mappings_model.index(last, 0)
        self._ui.source_list.selectionModel().setCurrentIndex(index, QItemSelectionModel.SelectionFlag.ClearAndSelect)

    @staticmethod
    def _is_url(string):
        """Tests if given string looks like a URL.

        Args:
            string (str): string to test

        Returns:
            bool: True if string looks like a URL, False otherwise
        """
        return "://" in string

    def tear_down(self):
        if not super().tear_down():
            return False
        if self._import_sources:
            self._import_sources.close_connection()
        return True


def _gams_system_directory(toolbox):
    """Returns GAMS system path from Toolbox settings or None if GAMS default is to be used."""
    path = toolbox.qsettings().value("appSettings/gamsPath", defaultValue=None)
    if not path:
        path = find_gams_directory()
    if path is not None and os.path.isfile(path):
        path = os.path.dirname(path)
    return path



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

""" Unit tests for :literal:`utils` module. """
from hashlib import sha1
from pathlib import Path
from tempfile import TemporaryDirectory
import unittest
from unittest import mock
from spine_engine.utils.helpers import AppSettings
from spine_items.tool.utils import find_last_output_files, get_julia_path_and_project


class TestFindLastOutputFiles(unittest.TestCase):
    def setUp(self):
        self._temp_dir = TemporaryDirectory()

    def tearDown(self):
        self._temp_dir.cleanup()

    def test_returns_empty_results_with_no_output_files_and_empty_output_dir(self):
        files = find_last_output_files([], self._temp_dir.name)
        self.assertEqual(files, {})

    def test_finds_output_file(self):
        stamped_dir = Path(self._temp_dir.name) / "2021-11-17T11.00.00"
        stamped_dir.mkdir()
        data_file = stamped_dir / "data.dat"
        data_file.touch()
        files = find_last_output_files(["data.dat"], self._temp_dir.name)
        self.assertEqual(files, {"data.dat": [str(data_file)]})

    def test_finds_output_file_in_subdirectory(self):
        stamped_dir = Path(self._temp_dir.name) / "2021-11-17T11.00.00"
        stamped_dir.mkdir()
        out_dir = stamped_dir / "out"
        out_dir.mkdir()
        data_file = out_dir / "data.dat"
        data_file.touch()
        files = find_last_output_files(["out/data.dat"], self._temp_dir.name)
        self.assertEqual(files, {"out/data.dat": [str(data_file)]})

    def test_finds_file_pattern(self):
        stamped_dir = Path(self._temp_dir.name) / "2021-11-17T11.00.00"
        stamped_dir.mkdir()
        data_file = stamped_dir / "data.dat"
        data_file.touch()
        files = find_last_output_files(["*.dat"], self._temp_dir.name)
        self.assertEqual(files, {"*.dat": [str(data_file)]})

    def test_finds_file_pattern_in_subdirectory(self):
        stamped_dir = Path(self._temp_dir.name) / "2021-11-17T11.00.00"
        stamped_dir.mkdir()
        out_dir = stamped_dir / "out"
        out_dir.mkdir()
        data_file = out_dir / "data.dat"
        data_file.touch()
        files = find_last_output_files(["out/*.dat"], self._temp_dir.name)
        self.assertEqual(files, {"out/*.dat": [str(data_file)]})

    def test_ignore_failed_directory(self):
        failed_dir = Path(self._temp_dir.name) / "failed"
        failed_dir.mkdir()
        data_file = failed_dir / "data.dat"
        data_file.touch()
        files = find_last_output_files(["data.dat"], self._temp_dir.name)
        self.assertEqual(files, {})

    def test_finds_latest_output_directory(self):
        stamped_dir_old = Path(self._temp_dir.name) / "2021-11-17T11.00.00"
        stamped_dir_old.mkdir()
        data_file_old = stamped_dir_old / "data.dat"
        data_file_old.touch()
        stamped_dir_new = Path(self._temp_dir.name) / "2021-11-17T11.30.00"
        stamped_dir_new.mkdir()
        data_file_new = stamped_dir_new / "data.dat"
        data_file_new.touch()
        files = find_last_output_files(["data.dat"], self._temp_dir.name)
        self.assertEqual(files, {"data.dat": [str(data_file_new)]})

    def test_finds_latest_output_directory_within_filter_id_directories(self):
        filter_id_dir = Path(self._temp_dir.name) / sha1(b"filter_id").hexdigest()
        filter_id_dir.mkdir()
        stamped_dir_old = filter_id_dir / "2021-11-17T11.00.00"
        stamped_dir_old.mkdir()
        data_file_old = stamped_dir_old / "data.dat"
        data_file_old.touch()
        stamped_dir_new = filter_id_dir / "2021-11-17T11.30.00"
        stamped_dir_new.mkdir()
        data_file_new = stamped_dir_new / "data.dat"
        data_file_new.touch()
        files = find_last_output_files(["data.dat"], self._temp_dir.name)
        self.assertEqual(files, {"data.dat": [str(data_file_new)]})

    def test_finds_latest_output_directory_within_filter_id_directories_with_mixed_directories(self):
        stamped_dir_old = Path(self._temp_dir.name) / "2021-11-17T11.00.00"
        stamped_dir_old.mkdir()
        data_file_old = stamped_dir_old / "data.dat"
        data_file_old.touch()
        filter_id_dir = Path(self._temp_dir.name) / sha1(b"filter_id").hexdigest()
        filter_id_dir.mkdir()
        stamped_dir_new = filter_id_dir / "2021-11-17T11.30.00"
        stamped_dir_new.mkdir()
        data_file_new = stamped_dir_new / "data.dat"
        data_file_new.touch()
        files = find_last_output_files(["data.dat"], self._temp_dir.name)
        self.assertEqual(files, {"data.dat": [str(data_file_new)]})

    def test_finds_latest_output_directory_within_non_filter_id_directories_with_mixed_directories(self):
        filter_id_dir = Path(self._temp_dir.name) / sha1(b"filter_id").hexdigest()
        filter_id_dir.mkdir()
        stamped_dir_old = filter_id_dir / "2021-11-17T11.00.00"
        stamped_dir_old.mkdir()
        data_file_old = stamped_dir_old / "data.dat"
        data_file_old.touch()
        stamped_dir_new = Path(self._temp_dir.name) / "2021-11-17T11.30.00"
        stamped_dir_new.mkdir()
        data_file_new = stamped_dir_new / "data.dat"
        data_file_new.touch()
        files = find_last_output_files(["data.dat"], self._temp_dir.name)
        self.assertEqual(files, {"data.dat": [str(data_file_new)]})


class TestGetJuliaPathAndProject(unittest.TestCase):
    def test_get_julia_path_and_project(self):
        # Use Jupyter Console: False
        exec_settings = {
            "use_jupyter_console": False,
            "executable": "",
            "env": "",
            "kernel_spec_name": "",
            "project": "",
        }
        app_settings = AppSettings({"appSettings/juliaPath": "path/to/julia"})
        julia_args = get_julia_path_and_project(exec_settings, app_settings)
        self.assertTrue(len(julia_args) == 1)
        self.assertTrue(julia_args[0] == "path/to/julia")
        # Use Jupyter Console: False
        exec_settings = {
            "use_jupyter_console": False,
            "executable": "/path/to/julia",
            "env": "",
            "kernel_spec_name": "",
            "project": "/path/to/myjuliaproject",
        }
        julia_args = get_julia_path_and_project(exec_settings, app_settings)
        self.assertTrue(julia_args[0] == "/path/to/julia")
        self.assertTrue(julia_args[1] == "--project=/path/to/myjuliaproject")
        # Use Jupyter Console: True
        exec_settings = {
            "use_jupyter_console": True,
            "executable": "",
            "env": "",
            "kernel_spec_name": "unknown_kernel",
            "project": "",
        }
        julia_args = get_julia_path_and_project(exec_settings, app_settings)
        self.assertIsNone(julia_args)
        # Use Jupyter Console: True
        exec_settings = {
            "use_jupyter_console": True,
            "executable": "/path/to/nowhere",
            "env": "conda",
            "kernel_spec_name": "test_kernel",
            "project": "/path/to/nonexistingprojectthatshouldnotbereturnedherebecausetheprojectisdefinedinkerneljson",
        }
        with mock.patch("spine_items.tool.utils.custom_find_kernel_specs") as mock_find_kernel_specs:
            # Return a dict containing a path to a dummy kernel resource dir when find_kernel_specs is called
            mock_find_kernel_specs.return_value = {"test_kernel": Path(__file__).parent / "dummy_julia_kernel"}
            julia_args = get_julia_path_and_project(exec_settings, app_settings)
            mock_find_kernel_specs.assert_called_once()
            self.assertEqual(2, len(julia_args))
            self.assertTrue(julia_args[0] == "/path/to/somejulia")
            self.assertTrue(julia_args[1] == "--project=/path/to/someotherjuliaproject")


if __name__ == "__main__":
    unittest.main()



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Toolbox.
# Spine Toolbox is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains a model to handle source tables and import mapping."""
from __future__ import annotations
from dataclasses import dataclass, field
from enum import IntEnum, unique
import pickle
import re
import uuid
from PySide6.QtCore import QAbstractItemModel, QMimeData, QModelIndex, Qt, Signal
from PySide6.QtGui import QColor, QFont
from spinedb_api import ParameterValueFormatError, from_database
from spinedb_api.import_mapping.import_mapping import default_import_mapping
from spinedb_api.import_mapping.import_mapping_compat import (
    import_mapping_from_dict,
    parameter_default_value_mapping_from_dict,
    parameter_mapping_from_dict,
    parameter_value_mapping_from_dict,
    parse_named_mapping_spec,
    unparse_named_mapping_spec,
)
from spinedb_api.import_mapping.type_conversion import DateTimeConvertSpec, FloatConvertSpec, StringConvertSpec
from spinedb_api.mapping import Position
from spinedb_api.mapping import to_dict as import_mapping_to_dict
from spinedb_api.parameter_value import join_value_and_type, split_value_and_type
from spinetoolbox.helpers import list_to_rich_text, plain_to_rich, unique_name
from spinetoolbox.mvcmodels.shared import PARSED_ROLE
from ..commands import (
    MoveMappingInList,
    MoveTableInList,
    PasteMappings,
    RenameMapping,
    SetFilterRe,
    SetMappingPosition,
    SetMappingPositionType,
    SetTableChecked,
    UpdateTableItem,
)
from ..flattened_mappings import VALUE_TYPES, FlattenedMappings
from ..mapping_colors import ERROR_COLOR
from ..widgets.mime_types import MAPPING_LIST_MIME_TYPE, SOURCE_TABLE_LIST_MIME_TYPE
from .mappings_model_roles import Role

UNNAMED_TABLE_NAME = "<rename this to add table>"


@unique
class FlattenedColumn(IntEnum):
    NAME = 0
    POSITION_TYPE = 1
    POSITION = 2
    REGEXP = 3


@dataclass()
class SourceTableItem:
    """A list item for source tables."""

    name: str
    checked: bool
    checkable: bool = True
    editable: bool = False
    accepts_drags_and_drops: bool = True
    real: bool = True
    in_source: bool = False
    in_specification: bool = False
    empty: bool = False
    select_all: bool = False
    mapping_list: list[MappingListItem] = field(init=False, default_factory=list)
    id: uuid.UUID = field(init=False, default_factory=uuid.uuid4)

    def append_to_mapping_list(self, list_item):
        """Appends an item to mapping list.

        Args:
            list_item (MappingListItem): item
        """
        list_item.source_table_item = self
        self.mapping_list.append(list_item)

    def insert_to_mapping_list(self, row, list_item):
        """Inserts an item to mapping list.

        Args:
            row (int): insertion position
            list_item (MappingListItem): item
        """
        list_item.source_table_item = self
        self.mapping_list.insert(row, list_item)


@dataclass
class SourceTableMimeData:
    # We currently support only internal move, so we need just the id here.
    source_table_id: uuid.UUID


@dataclass
class MappingMimeData:
    name: str
    mapping_id: uuid.UUID
    mapping_dict: list[dict]


@dataclass
class MappingListMimeData:
    """Mime data for dragging and dropping."""

    source_table_name: str
    source_table_id: uuid.UUID
    mapping_list: list[MappingMimeData]


_dummy = object()


@dataclass()
class MappingListItem:
    """An item in mappings list."""

    name: str
    flattened_mappings: FlattenedMappings = field(init=False)
    source_table_item: SourceTableItem = field(init=False)
    id: uuid.UUID = field(init=False, default_factory=uuid.uuid4)

    def set_flattened_mappings(self, flattened_mappings):
        """Sets flattened mappings for this item.

        Args:
            flattened_mappings (FlattenedMappings): mappings
        """
        flattened_mappings.mapping_list_item = self
        self.flattened_mappings = flattened_mappings


class MappingsModel(QAbstractItemModel):
    """A tree model that manages source tables and their mappings."""

    msg_error = Signal(str)
    """Emitted when an error message should be displayed."""
    row_or_column_type_recommended = Signal(int, object, object)
    """Emitted when a change in mapping prompts for change in column or row type."""
    multi_column_type_recommended = Signal(object, object)
    """Emitted when all but given columns should be of given type."""

    _NOT_IN_SOURCE_COLOR = QColor(Qt.GlobalColor.gray)

    def __init__(self, undo_stack, parent):
        """
        Args:
            undo_stack (QUndoStack): undo stack
            parent (QObject): parent object
        """
        super().__init__(parent)
        self._undo_stack = undo_stack
        self._mappings = [self._make_select_all_tables_item()]
        self._add_table_row_font = QFont()
        self._add_table_row_font.setItalic(True)

    def __len__(self):
        return len(self._mappings)

    def has_empty_source_table_row(self):
        """Tests if source table list contains the empty row.

        Returns:
            bool: True if the last row is empty table, False otherwise
        """
        return self._mappings[-1].empty

    def real_table_names(self):
        """Returns real table names.

        Returns:
            list of str: real table names
        """
        return [i.name for i in self._mappings if i.real]

    def checked_table_names(self):
        """Returns checked table names.

        Returns:
            list of str: checked table names
        """
        return [i.name for i in self._mappings[1:] if i.checked]

    @staticmethod
    def _make_select_all_tables_item():
        """Creates the first source table list item.

        Returns:
            SourceTableItem: 'select all' item
        """
        return SourceTableItem("Select all", checked=True, accepts_drags_and_drops=False, real=False, select_all=True)

    def columnCount(self, parent=QModelIndex()):
        if not parent.isValid():
            return 1
        parent_item = parent.internalPointer()
        if isinstance(parent_item, SourceTableItem):
            return 1
        if isinstance(parent_item, MappingListItem):
            return 4
        return 0

    def data(self, index, role=Qt.ItemDataRole.DisplayRole):
        index_item = index.internalPointer()
        if isinstance(index_item, SourceTableItem):
            return self._table_list_data(index, role)
        if isinstance(index_item, MappingListItem):
            return self._mapping_list_data(index_item, role)
        if isinstance(index_item, FlattenedMappings):
            return self._mapping_data(index_item, index, role)
        return None

    def _table_list_data(self, index, role):
        """Returns source table data for given role.

        Args:
            index (QModelIndex): table index
            role (int): Qt's data role

        Returns:
            Any: data
        """
        if role == Qt.ItemDataRole.DisplayRole:
            row = index.row()
            list_item = self._mappings[row]
            if row == 0 or list_item.in_specification or list_item.empty:
                return list_item.name
            return list_item.name + " (new)"
        if role == Qt.ItemDataRole.EditRole:
            return self._mappings[index.row()].name
        if role == Qt.ItemDataRole.CheckStateRole:
            return Qt.CheckState.Checked if self._mappings[index.row()].checked else Qt.CheckState.Unchecked
        if role == Qt.ItemDataRole.ForegroundRole:
            row = index.row()
            item = self._mappings[row]
            return self._NOT_IN_SOURCE_COLOR if row > 0 and (not item.in_source and not item.empty) else None
        if role == Qt.ItemDataRole.ToolTipRole:
            row = index.row()
            if row == 0:
                return None
            list_item = self._mappings[row]
            if not list_item.empty:
                if not list_item.in_source:
                    return plain_to_rich("Table isn't in source data.")
                if not list_item.in_specification:
                    return plain_to_rich("Table's mappings haven't been saved with the specification yet.")
            return None
        if role == Qt.ItemDataRole.FontRole:
            return self._add_table_row_font if self._mappings[index.row()].empty else None
        if role == Role.ITEM:
            return self._mappings[index.row()]
        return None

    @staticmethod
    def _mapping_list_data(list_item, role):
        """Returns mapping list data for given role.

        Args:
            list_item (MappingListItem): list item
            role (int): Qt's data role

        Returns:
            Any: data
        """
        if role in (Qt.ItemDataRole.DisplayRole, Qt.ItemDataRole.EditRole):
            return list_item.name
        if role == Role.ITEM:
            return list_item
        if role == Role.FLATTENED_MAPPINGS:
            return list_item.flattened_mappings
        return None

    @staticmethod
    def _mapping_data(flattened_mappings, index, role):
        """Returns mapping component data for given role.

        Args:
            flattened_mappings (FlattenedMappings): flattened mappings
            index (QModelIndex): component index
            role (int): Qt's data role

        Returns:
            Any: data
        """
        if role == PARSED_ROLE:
            # Only used for the ParameterValueEditor.
            # At this point, the delegate has already checked that it's a constant parameter (default) value mapping
            m = flattened_mappings.component_at(index.row())
            try:
                return from_database(*split_value_and_type(m.value))
            except ParameterValueFormatError:
                return None
        column = index.column()
        if role in (Qt.ItemDataRole.DisplayRole, Qt.ItemDataRole.EditRole):
            if column == FlattenedColumn.NAME:
                return flattened_mappings.display_names[index.row()]
            if column == FlattenedColumn.POSITION_TYPE:
                return flattened_mappings.display_position_type(index.row())
            if column == FlattenedColumn.POSITION:
                display = flattened_mappings.display_position(index.row())
                if display == "<table name>":
                    return flattened_mappings.mapping_list_item.source_table_item.name
                if display == "<mapping name>":
                    return flattened_mappings.mapping_list_item.name
                return display
            if column == FlattenedColumn.REGEXP:
                return flattened_mappings.component_at(index.row()).filter_re
            raise RuntimeError("Column out of bounds.")
        if role == Qt.ItemDataRole.BackgroundRole:
            if column == FlattenedColumn.NAME:
                return flattened_mappings.display_colors[index.row()]
            if column == FlattenedColumn.POSITION:
                issues = flattened_mappings.display_row_issues(index.row())
                if issues:
                    return ERROR_COLOR
                return None
        if role == Qt.ItemDataRole.ToolTipRole:
            if column == FlattenedColumn.POSITION:
                issues = flattened_mappings.display_row_issues(index.row())
                return list_to_rich_text(issues) if issues else None
            if column == FlattenedColumn.REGEXP:
                return plain_to_rich("Enter regular expression to filter importer data.")
        return None

    def headerData(self, section, orientation, role=Qt.ItemDataRole.DisplayRole):
        if role == Qt.ItemDataRole.DisplayRole and orientation == Qt.Orientation.Horizontal:
            return ("Target", "Source type", "Source ref.", "Filter")[section]
        return None

    def flags(self, index):
        index_item = index.internalPointer()
        if index_item is None:
            return super().flags(index) | Qt.ItemFlag.ItemIsDropEnabled
        if isinstance(index_item, SourceTableItem):
            return self._table_list_flags(index)
        if isinstance(index_item, MappingListItem):
            return self._mapping_list_flags()
        if isinstance(index_item, FlattenedMappings):
            return self._mapping_flags(index_item, index)
        return super().flags(index)

    def _table_list_flags(self, index):
        """Returns flags for source table list.

        Args:
            index (QModelIndex): index

        Returns:
            int: flags
        """
        flags = Qt.ItemFlag.ItemIsEnabled
        table_item = self._mappings[index.row()]
        if table_item.checkable:
            flags |= Qt.ItemFlag.ItemIsUserCheckable
        if table_item.editable:
            flags |= Qt.ItemFlag.ItemIsEditable
        if table_item.accepts_drags_and_drops:
            flags |= Qt.ItemFlag.ItemIsDragEnabled
            flags |= Qt.ItemFlag.ItemIsDropEnabled
        if not table_item.select_all:
            flags |= Qt.ItemFlag.ItemIsSelectable
        return flags

    @staticmethod
    def _mapping_list_flags():
        """Returns flags for mapping list data.

        Returns:
            Qt.ItemFlag: flags
        """
        return (
            Qt.ItemFlag.ItemIsSelectable
            | Qt.ItemFlag.ItemIsEnabled
            | Qt.ItemFlag.ItemIsEditable
            | Qt.ItemFlag.ItemIsDragEnabled
        )

    @staticmethod
    def _mapping_flags(flattened_item, index):
        """Returns flags for mapping components.

        Args:
            flattened_item (FlattenedMappings): flattened mappings
            index (QModelIndex): index

        Returns:
            int: flags
        """
        non_editable = Qt.ItemIsEnabled | Qt.ItemIsSelectable
        editable = non_editable | Qt.ItemIsEditable
        column = index.column()
        if column == FlattenedColumn.NAME:
            return non_editable
        row = index.row()
        if flattened_item.is_component_pivoted(row):
            return non_editable
        if column == FlattenedColumn.POSITION:
            component = flattened_item.component_at(index.row())
            if component.value is None and component.position == Position.header:
                return non_editable
        return editable

    def mimeTypes(self):
        return [SOURCE_TABLE_LIST_MIME_TYPE, MAPPING_LIST_MIME_TYPE]

    def mimeData(self, indexes):
        if not indexes:
            return QMimeData()
        items = [i.internalPointer() for i in indexes]
        if isinstance(items[0], SourceTableItem):
            mime_type = SOURCE_TABLE_LIST_MIME_TYPE
            if not all(isinstance(item, SourceTableItem) for item in items[1:]):
                raise RuntimeError("Logic error: not all indexes are SourceTableItems.")
            data = self._source_table_list_mime_data(items)
        elif isinstance(items[0], MappingListItem):
            mime_type = MAPPING_LIST_MIME_TYPE
            source_table_name = items[0].source_table_item.name
            source_table_id = items[0].source_table_item.id
            if not all(isinstance(item, MappingListItem) for item in items[1:]):
                raise RuntimeError("Logic error: not all indexes are MappingListItems.")
            if not all(item.source_table_item.id == source_table_id for item in items[1:]):
                raise RuntimeError("Logic error: not all MappingListItems come from same source table.")
            data = self._mapping_list_mime_data(source_table_name, source_table_id, items)
        else:
            raise RuntimeError("Logic error: drag and drop not supported for given indexes")
        mime_data = QMimeData()
        mime_data.setData(mime_type, pickle.dumps(data))
        return mime_data

    @staticmethod
    def _source_table_list_mime_data(items):
        """Creates MIME data for source table items.

        Args:
            items (Iterable of SourceTableItem): items

        Returns:
            MappingListMimeData: MIME data
        """

        data = []
        for item in items:
            data.append(SourceTableMimeData(item.id))
        return data

    @staticmethod
    def _mapping_list_mime_data(source_table_name, source_table_id, items):
        """Creates MIME data for mapping items that are all from a single source table.

        Args:
            source_table_name (str): name of the source table
            source_table_id (uuid.UUID): id of the source table
            items (Iterable of MappingListItem): items

        Returns:
            MappingListMimeData: MIME data
        """
        mapping_mime_data_list = []
        for item in items:
            mapping_mime_data_list.append(
                MappingMimeData(
                    item.name,
                    item.id,
                    import_mapping_to_dict(item.flattened_mappings.root_mapping),
                )
            )
        return MappingListMimeData(source_table_name, source_table_id, mapping_mime_data_list)

    def canDropMimeData(self, data, action, row, column, parent):
        if not any(data.hasFormat(mime_type) for mime_type in self.mimeTypes()):
            return False
        parent_item = parent.internalPointer()
        if data.hasFormat(SOURCE_TABLE_LIST_MIME_TYPE):
            if parent_item is not None:
                return False
            if row == 0:
                return False
            if self.has_empty_source_table_row() and (row == -1 or row == len(self._mappings)):
                return False
            return True
        if not isinstance(parent_item, SourceTableItem):
            return False
        mapping_list_data: MappingListMimeData = pickle.loads(data.data(MAPPING_LIST_MIME_TYPE))
        source_table: SourceTableItem = parent.internalPointer()
        is_internal_move = source_table.id == mapping_list_data.source_table_id
        if is_internal_move and row == -1:
            return False
        return True

    def dropMimeData(self, data, action, row, column, parent):
        if not self.canDropMimeData(data, action, row, column, parent):
            return False
        if data.hasFormat(SOURCE_TABLE_LIST_MIME_TYPE):
            return self._drop_source_table_list(data, row)
        if data.hasFormat(MAPPING_LIST_MIME_TYPE):
            return self._drop_mappings_list(data, row, parent)
        raise RuntimeError(f"Logic error: unknown mime types {data.formats()}")

    def _drop_source_table_list(self, data, row):
        """Drops source table list.

        Args:
            data (QMimeData): MIME data
            row (int): target row index

        Returns:
            bool: True if operation was successful, False otherwise
        """
        source_table_list_data: list[SourceTableMimeData] = pickle.loads(data.data(SOURCE_TABLE_LIST_MIME_TYPE))
        moves = []
        # We support internal moves only currently.
        while source_table_list_data:
            source_table_data = source_table_list_data.pop(0)
            for i, source_table in enumerate(self._mappings[1:]):
                if source_table.id == source_table_data.source_table_id:
                    table_row = i + 1
                    # Ensure we're not dropping on the same spot.
                    if table_row not in {row, row - 1}:
                        moved_from_above = len(
                            [source for source, destination in moves if source < table_row < destination]
                        )
                        moves.append((table_row - moved_from_above, row))
                        row = row if row > table_row else row + 1
                    break
        if not moves:
            return False
        self._undo_stack.beginMacro("change the order of tables")
        for source_row, target_row in moves:
            self._undo_stack.push(MoveTableInList(self, source_row, target_row))
        self._undo_stack.endMacro()
        return True

    def _drop_mappings_list(self, data, row, parent):
        """Drops mapping list.

        Args:
            data (QMimeData): MIME data
            row (int): target row index
            parent (QModelIndex): target index

        Returns:
            bool: True if operation was successful, False otherwise
        """
        source_table: SourceTableItem = parent.internalPointer()
        mapping_list_data: MappingListMimeData = pickle.loads(data.data(MAPPING_LIST_MIME_TYPE))
        is_internal_move = source_table.id == mapping_list_data.source_table_id
        if is_internal_move:
            moves = []
            while mapping_list_data.mapping_list:
                mapping_data = mapping_list_data.mapping_list.pop(0)
                for i, mapping in enumerate(source_table.mapping_list):
                    if mapping.id == mapping_data.mapping_id:
                        # Ensure we're not dropping on the same spot.
                        if i not in {row, row - 1}:
                            moved_from_above = len(
                                [source for source, destination in moves if source < i < destination]
                            )
                            moves.append((i - moved_from_above, row))
                            row = row if row > i else row + 1
                        break
            if not moves:
                return False
            self._undo_stack.beginMacro("change the order of mappings")
            source_table_row = parent.row()
            for source_row, target_row in moves:
                self._undo_stack.push(MoveMappingInList(source_table_row, self, source_row, target_row))
            self._undo_stack.endMacro()
        else:
            if row == -1:
                row = len(source_table.mapping_list)
            self._undo_stack.push(PasteMappings(parent.row(), row, self, mapping_list_data, "drop mappings"))
        return True

    def insert_mapping_mime_data(self, table_row, list_row, mapping_list_data):
        """Inserts mapping mime data to mapping list

        Args:
            table_row (int): index to source table list
            list_row (int): index to mapping list
            mapping_list_data (list of MappingMimeData): mappings to insert
        """
        parent = self.index(table_row, 0)
        source_table: SourceTableItem = parent.internalPointer()
        self.beginInsertRows(parent, list_row, list_row + len(mapping_list_data) - 1)
        for i, mapping_data in enumerate(mapping_list_data):
            mapping_item = MappingListItem(mapping_data.name)
            flattened_mappings = FlattenedMappings(import_mapping_from_dict(mapping_data.mapping_dict))
            mapping_item.set_flattened_mappings(flattened_mappings)
            source_table.insert_to_mapping_list(list_row + i, mapping_item)
        self.endInsertRows()

    def index(self, row, column, parent=QModelIndex()):
        if row < 0 or column < 0:
            return QModelIndex()
        if not parent.isValid():
            return self.createIndex(row, 0, self._mappings[row])
        parent_item = parent.internalPointer()
        if isinstance(parent_item, SourceTableItem):
            return self.createIndex(row, 0, parent_item.mapping_list[row])
        if isinstance(parent_item, MappingListItem):
            return self.createIndex(row, column, parent_item.flattened_mappings)
        raise RuntimeError("Cannot create index.")

    @staticmethod
    def is_table_index(index):
        """Checks if given index is an index to source table list.

        Args:
            index (QModelIndex): index

        Returns:
            bool: True if index is table index, False otherwise
        """
        return isinstance(index.internalPointer(), SourceTableItem)

    @staticmethod
    def is_mapping_list_index(index):
        """Checks if given index is an index to mapping list.

        Args:
            index (QModelIndex): index

        Returns:
            bool: True if index is mapping list index, False otherwise
        """
        return isinstance(index.internalPointer(), MappingListItem)

    def dummy_parent(self):
        """Creates a dummy parent index that always returns an empty table.

        Returns:
            QModelIndex: dummy index
        """
        return self.createIndex(0, 0, _dummy)

    def parent(self, index):
        index_item = index.internalPointer()
        if isinstance(index_item, SourceTableItem):
            return QModelIndex()
        if isinstance(index_item, MappingListItem):
            for table_row, table_item in enumerate(self._mappings):
                if index_item.source_table_item is table_item:
                    return self.createIndex(table_row, 0, table_item)
        elif isinstance(index_item, FlattenedMappings):
            for list_row, list_item in enumerate(index_item.mapping_list_item.source_table_item.mapping_list):
                if index_item is list_item.flattened_mappings:
                    return self.createIndex(list_row, 0, list_item)
        elif index_item is _dummy:
            return QModelIndex()
        raise RuntimeError("Cannot create parent index.")

    def hasChildren(self, parent=QModelIndex()):
        if not QModelIndex().isValid():
            return True
        index_item = parent.internalPointer()
        if isinstance(index_item, SourceTableItem):
            return bool(index_item.mapping_list)
        if isinstance(index_item, MappingListItem):
            return True
        return False

    def store(self):
        """Stores source tables, mappings lists and mappings to a dict.

        Returns:
            dict: stored data
        """
        table_dicts = {}
        for table_item in self._mappings[1:]:
            if not table_item.real:
                continue
            table_dicts[table_item.name] = [
                unparse_named_mapping_spec(list_item.name, list_item.flattened_mappings.root_mapping)
                for list_item in table_item.mapping_list
            ]
        selected_tables = self.checked_table_names()
        return {"table_mappings": table_dicts, "selected_tables": selected_tables}

    def restore(self, mappings_dict):
        """Restores source tables, mapping lists and mappings.

        Args:
            mappings_dict (dict): serialized data
        """
        checked_tables = mappings_dict.get("selected_tables")
        self.beginResetModel()
        self._mappings.clear()
        self._mappings.append(self._make_select_all_tables_item())
        try:
            for table_name, mapping_list_dicts in mappings_dict.get("table_mappings", {}).items():
                checked = table_name in checked_tables if checked_tables is not None else True
                table_item = SourceTableItem(table_name, checked, in_specification=True)
                for list_dict in mapping_list_dicts:
                    mapping_name, root_mapping = parse_named_mapping_spec(list_dict)
                    flattened_mappings = FlattenedMappings(root_mapping)
                    list_item = MappingListItem(mapping_name)
                    list_item.set_flattened_mappings(flattened_mappings)
                    table_item.append_to_mapping_list(list_item)
                self._mappings.append(table_item)
        except ValueError as error:
            # This will be raised by parse_named_mapping_spec() if mapping type is obsolete, like tool, feature, etc.
            self.msg_error.emit(f"{error}")
            return
        self.endResetModel()
        self._update_all_checked()

    def rowCount(self, parent=QModelIndex()):
        if not parent.isValid():
            return len(self._mappings)
        parent_item = parent.internalPointer()
        if isinstance(parent_item, SourceTableItem):
            return len(parent_item.mapping_list)
        if isinstance(parent_item, MappingListItem):
            if parent_item.flattened_mappings is None:
                return 0
            return len(parent_item.flattened_mappings.display_names)
        return 0

    def setData(self, index, value, role=Qt.ItemDataRole.EditRole):
        index_item = index.internalPointer()
        if isinstance(index_item, SourceTableItem):
            return self._set_table_list_data(index_item, index, value, role)
        if isinstance(index_item, MappingListItem):
            return self._set_mapping_list_data(index_item, index, value, role)
        if isinstance(index_item, FlattenedMappings):
            return self._set_mapping_data(index_item, index, value, role)
        raise RuntimeError("Unknown parent type.")

    def _set_table_list_data(self, table_item, index, value, role):
        """Sets data for source table list.

        Args:
            table_item (SourceTableItem): table item
            index (QModelIndex): index
            value (Any): value to set
            role (int): Qt's data role

        Returns:
            bool: True if data was set successfully, False otherwise
        """
        row = index.row()
        if role == Qt.ItemDataRole.CheckStateRole and table_item.checkable:
            checked = value == Qt.CheckState.Checked.value
            if row == 0:
                self._set_multiple_checked_undoable(checked, *range(len(self._mappings)))
            else:
                self._undo_stack.push(SetTableChecked(table_item.name, self, checked, row))
            return True
        if role == Qt.ItemDataRole.EditRole:
            if value == table_item.name:
                return False
            if value in {table.name for table in self._mappings if table.real}:
                self.msg_error.emit(f"There's already a table called {value}.")
                return False
            previous = {"name": table_item.name}
            updates = {"name": value}
            self._undo_stack.push(UpdateTableItem(self, row, previous, updates, table_item.empty))
            return True
        return False

    def _set_mapping_list_data(self, list_item, index, value, role):
        """Sets data for mapping list.

        Args:
            list_item (MappingListItem): list item
            index (QModelIndex): index
            value (Any): value to set
            role (int): Qt's data role

        Returns:
            bool: True if data was set successfully, False otherwise
        """
        if role == Qt.ItemDataRole.EditRole:
            if not value or value in {m.name for m in list_item.source_table_item.mapping_list}:
                return False
            previous = list_item.name
            parent_index = index.parent()
            self._undo_stack.push(RenameMapping(parent_index.row(), index.row(), self, value, previous))
            return True
        return False

    def update_table_item(self, row, data_dict, add_empty_row=False, remove_empty_row=False):
        """Updates table item.

         This only happens in file-less mode, when the user is creating tables.

        Args:
            row (int): item row
            data_dict (dict): table item data to update
            add_empty_row (bool): whether to add an empty row
            remove_empty_row (bool): whether to remove the empty row
        """
        table_item = self._mappings[row]
        for field_name, value in data_dict.items():
            setattr(table_item, field_name, value)
        index = self.index(row, 0)
        if add_empty_row:
            table_item.real = True
            table_item.checkable = True
            table_item.checked = self._mappings[0].checked
            table_item.accepts_drags_and_drops = True
            table_item.empty = False
            table_item.in_source = True
            default_flattened_mappings = FlattenedMappings(self.create_default_mapping())
            mapping_list_item = MappingListItem(self._unique_mapping_name(table_item))
            mapping_list_item.set_flattened_mappings(default_flattened_mappings)
            self.beginInsertRows(index, 0, 0)
            table_item.append_to_mapping_list(mapping_list_item)
            self.endInsertRows()
            self.add_empty_row()
        if remove_empty_row:
            table_item.real = False
            table_item.checkable = False
            table_item.checked = False
            table_item.accepts_drags_and_drops = False
            table_item.empty = True
            table_item.in_source = False
            self.beginRemoveRows(index, 0, len(table_item.mapping_list) - 1)
            table_item.mapping_list.clear()
            self.endRemoveRows()
            self._remove_last_source_table()
        self.dataChanged.emit(index, index)

    def toggle_checked_tables(self, table_indexes):
        """Toggles the checked status of source tables.

        Args:
            table_indexes (Iterable of QModelIndex): indexes
        """
        table_items = ((i.row(), self._mappings[i.row()]) for i in table_indexes)
        checkables = [i for i in table_items if i[1].checkable and i[1].real]
        all_checked = all(i[1].checked for i in checkables)
        self._set_multiple_checked_undoable(not all_checked, *(i[0] for i in checkables))

    def _set_multiple_checked_undoable(self, checked, *rows):
        """
        Sets the checked status of multiple list items.

        This action is undoable.

        Args:
            checked (bool): True for checked, False for unchecked
            *rows: item rows
        """
        rows = [row for row in rows if self._mappings[row].checked != checked]
        self._undo_stack.push(SetTableChecked(None, self, checked, *rows))

    def set_table_checked(self, checked, *rows):
        """
        Sets the checked status of a source table item.

        Args:
            checked (bool): True for checked, False for unchecked
            *rows: item rows
        """
        min_row = None
        max_row = None
        for row in rows:
            if not self._mappings[row].checkable or self._mappings[row].checked == checked:
                continue
            if min_row is None:
                min_row = row
                max_row = row
            else:
                min_row = min(row, min_row)
                max_row = max(row, max_row)
            self._mappings[row].checked = checked
        if min_row is not None:
            top_left = self.index(min_row, 0)
            bottom_right = self.index(max_row, 0)
            self.dataChanged.emit(top_left, bottom_right, [Qt.ItemDataRole.CheckStateRole])
        self._update_all_checked()

    def _update_all_checked(self):
        """Updates the checked state of 'Select All' table item if needed."""
        checkables = tuple(m for m in self._mappings[1:] if m.checkable)
        if not checkables:
            return
        all_checked = all(m.checked for m in checkables)
        all_checked_item = self._mappings[0]
        if all_checked_item.checked != all_checked:
            self._mappings[0].checked = all_checked
            index = self.index(0, 0)
            self.dataChanged.emit(index, index, [Qt.ItemDataRole.CheckStateRole])

    def append_new_table_with_mapping(self, table_name, root_mapping):
        """Appends a new source table and adds given ``root_mapping`` as its first mapping.

        Args:
            table_name (str): table name
            root_mapping (ImportMapping, optional): initial mapping
        """
        has_root_mapping = root_mapping is not None
        if not has_root_mapping:
            root_mapping = self.create_default_mapping()
        flattened_mappings = FlattenedMappings(root_mapping)
        list_item = MappingListItem("Mapping 1")
        list_item.set_flattened_mappings(flattened_mappings)
        table_item = SourceTableItem(
            table_name, checked=self._mappings[0].checked, in_source=True, in_specification=has_root_mapping
        )
        table_item.append_to_mapping_list(list_item)
        insertion_point = len(self._mappings) - (1 if self.has_empty_source_table_row() else 0)
        self.beginInsertRows(QModelIndex(), insertion_point, insertion_point)
        self._mappings.insert(insertion_point, table_item)
        self.endInsertRows()

    def _remove_last_source_table(self):
        """Removes the last source table."""
        self.beginRemoveRows(QModelIndex(), self.rowCount() - 1, self.rowCount() - 1)
        del self._mappings[-1]
        self.endRemoveRows()

    def rename_mapping(self, table_row, list_row, name):
        """
        Renames a mapping.

        Args:
            table_row (int): source table row
            list_row (int): mapping list row
            name (str): new name
        """
        table_item = self._mappings[table_row]
        list_item = table_item.mapping_list[list_row]
        if name == list_item.name:
            return
        list_item.name = ""
        if self.has_mapping_name(table_row, name):
            match = re.search(r" [0-9]+$", name)
            prefix = name[: match.start()] if match is not None else name
            name = self._unique_mapping_name(table_item, prefix)
        list_item.name = name
        parent = self.index(table_row, 0)
        index = self.index(list_row, 0, parent)
        self.dataChanged.emit(index, index, [Qt.ItemDataRole.DisplayRole])

    def set_root_mapping(self, table_row, list_row, root_mapping):
        """
        Sets new root mapping.

        Args:
            table_row (int): source table row index
            list_row (int): mapping list row index
            root_mapping (ImportMapping): root mapping
        """
        table_index = self.index(table_row, 0)
        list_index = self.index(list_row, 0, table_index)
        mapping_list = self._mappings[table_row].mapping_list[list_row]
        flattened_mappings = mapping_list.flattened_mappings
        self.beginRemoveRows(list_index, 0, len(flattened_mappings.display_names) - 1)
        mapping_list.flattened_mappings = None
        self.endRemoveRows()
        flattened_mappings.set_root_mapping(root_mapping)
        self.beginInsertRows(list_index, 0, len(flattened_mappings.display_names) - 1)
        mapping_list.flattened_mappings = flattened_mappings
        self.endInsertRows()
        self.dataChanged.emit(list_index, list_index, [Role.FLATTENED_MAPPINGS])

    def set_mappings_type(self, table_row, list_row, new_type):
        """
        Sets the type for mappings.

        Args:
            table_row (int): source table row index
            list_row (int): mapping list row index
            new_type (str): name of the type
        """
        map_type = {
            "Entity class": "EntityClass",
            "Entity group": "EntityGroup",
            "Alternative": "Alternative",
            "Scenario": "Scenario",
            "Scenario alternative": "ScenarioAlternative",
            "Parameter value list": "ParameterValueList",
        }[new_type]
        root_mapping = default_import_mapping(map_type)
        self.set_root_mapping(table_row, list_row, root_mapping)

    def _set_mapping_data(self, flattened_mappings, index, value, role):
        """Sets data for mapping components.

        Args:
            flattened_mappings (FlattenedMappings): component's mapping list
            index (QModelIndex): index
            value (Any): value to set
            role (int): Qt's data role

        Returns:
            bool: True if data was set successfully, False otherwise
        """
        if role != Qt.ItemDataRole.EditRole:
            return False
        column = index.column()
        if column < FlattenedColumn.POSITION_TYPE:
            return False
        row = index.row()
        current_type = flattened_mappings.display_position_type(row)
        current_position = flattened_mappings.display_position(row)
        if column == FlattenedColumn.POSITION_TYPE:
            if value == current_type:
                return False
            parent_index = index.parent()
            mapping_list_row = parent_index.row()
            table_row = parent_index.parent().row()
            self._undo_stack.push(
                SetMappingPositionType(table_row, mapping_list_row, row, self, value, current_type, current_position)
            )
            return True
        if column == FlattenedColumn.POSITION:
            if current_type != "None":
                if value == current_position:
                    return False
                parent_index = index.parent()
                mapping_list_row = parent_index.row()
                table_row = parent_index.parent().row()
                self._undo_stack.push(
                    SetMappingPosition(
                        table_row, mapping_list_row, row, self, current_type, value, current_type, current_position
                    )
                )
                return True
            # If type is "None", set it to something reasonable to try and help users
            try:
                value = int(value)
            except ValueError:
                pass
            if isinstance(value, int):
                return self.change_component_mapping(flattened_mappings, index, "Column", value)
            if isinstance(value, str):
                return self.change_component_mapping(flattened_mappings, index, "Constant", value)
        if column == FlattenedColumn.REGEXP:
            parent_index = index.parent()
            mapping_list_row = parent_index.row()
            table_row = parent_index.parent().row()
            current_re = flattened_mappings.component_at(row).filter_re
            self._undo_stack.push(SetFilterRe(table_row, mapping_list_row, row, self, value, current_re))
            return True
        return False

    def get_set_data_delayed(self, index):
        """Returns a function that ParameterValueEditor can call to set data for the given index at any later time,
        even if the model changes.

        Args:
            index (QModelIndex): model index

        Returns:
            Callable: callback for parameter value editor
        """
        return lambda value_type_tup, index=index: self.setData(index, join_value_and_type(*value_type_tup))

    @staticmethod
    def index_name(index):
        """Returns identifier for parameter value editor's title.

        Returns:
            str: 'index' name
        """
        return index.siblingAtColumn(0).data()

    def set_relationship_dimension_count(self, table_row, list_row, dimension_count):
        """Sets relationship dimension count.

        Args:
            table_row (int): source table row index
            list_row (int): mapping list row index
            dimension_count (int): dimension count
        """
        table_index = self.index(table_row, 0)
        list_index = self.index(list_row, 0, table_index)
        mapping_list = self._mappings[table_row].mapping_list[list_row]
        flattened_mappings = mapping_list.flattened_mappings
        self.beginRemoveRows(list_index, 0, len(flattened_mappings.display_names) - 1)
        mapping_list.flattened_mappings = None
        self.endRemoveRows()
        flattened_mappings.set_dimension_count(dimension_count)
        self.beginInsertRows(list_index, 0, len(flattened_mappings.display_names) - 1)
        mapping_list.flattened_mappings = flattened_mappings
        self.endInsertRows()
        self.dataChanged.emit(list_index, list_index, [Role.FLATTENED_MAPPINGS])

    def set_map_dimension_count(self, table_row, list_row, dimension_count):
        """
        Sets map dimension_count.

        Args:
            table_row (int): source table row index
            list_row (int): mapping list row index
            dimension_count (int): new map dimension_count
        """
        table_index = self.index(table_row, 0)
        list_index = self.index(list_row, 0, table_index)
        mapping_list = self._mappings[table_row].mapping_list[list_row]
        flattened_mappings = mapping_list.flattened_mappings
        self.beginRemoveRows(list_index, 0, len(flattened_mappings.display_names) - 1)
        mapping_list.flattened_mappings = None
        self.endRemoveRows()
        flattened_mappings.set_map_dimension_count(dimension_count)
        self.beginInsertRows(list_index, 0, len(flattened_mappings.display_names) - 1)
        mapping_list.flattened_mappings = flattened_mappings
        self.endInsertRows()
        self.dataChanged.emit(list_index, list_index, [Role.FLATTENED_MAPPINGS])

    def set_map_compress(self, table_row, list_row, compress):
        """
        Sets the compress flag for Map type parameters.

        Args:
            table_row (int): source table row index
            list_row (int): mapping list row index
            compress (bool): flag value
        """
        flattened_mappings = self._mappings[table_row].mapping_list[list_row].flattened_mappings
        flattened_mappings.set_map_compress(compress)
        table_index = self.index(table_row, 0)
        list_index = self.index(list_row, 0, table_index)
        self.dataChanged.emit(list_index, list_index, [Role.FLATTENED_MAPPINGS])

    def set_parameter_type(self, table_row, list_row, new_type):
        """Changes parameter type.

        Args:
            table_row (int): source table row index
            list_row (int): mapping list row index
            new_type (str): parameter type
        """
        if new_type == "None":
            parameter_root = None
        elif new_type == "Value":
            parameter_root = parameter_mapping_from_dict({"map_type": "ParameterValue"})
        elif new_type == "Definition":
            parameter_root = parameter_mapping_from_dict({"map_type": "ParameterDefinition"})
        else:
            raise RuntimeError(f"Unknown parameter type '{new_type}'")
        table_index = self.index(table_row, 0)
        list_index = self.index(list_row, 0, table_index)
        mapping_list = self._mappings[table_row].mapping_list[list_row]
        flattened_mappings = mapping_list.flattened_mappings
        self.beginRemoveRows(list_index, 0, len(flattened_mappings.display_names) - 1)
        mapping_list.flattened_mappings = None
        self.endRemoveRows()
        flattened_mappings.set_parameter_components(parameter_root)
        self.beginInsertRows(list_index, 0, len(flattened_mappings.display_names) - 1)
        mapping_list.flattened_mappings = flattened_mappings
        self.endInsertRows()
        self.dataChanged.emit(list_index, list_index, [Role.FLATTENED_MAPPINGS])

    def set_value_type(self, table_row, list_row, new_type):
        """Changes parameter value's type.

        Args:
            table_row (int): source table row index
            list_row (int): mapping list row index
            new_type (str): value type
        """
        table_index = self.index(table_row, 0)
        list_index = self.index(list_row, 0, table_index)
        mapping_list = self._mappings[table_row].mapping_list[list_row]
        flattened_mappings = mapping_list.flattened_mappings
        if new_type == "None":
            value_mappings = None
        else:
            value_type = VALUE_TYPES[new_type]
            parameter_type = flattened_mappings.display_parameter_type()
            if parameter_type == "Definition":
                value_mappings = parameter_default_value_mapping_from_dict({"value_type": value_type})
            elif parameter_type == "Value":
                value_mappings = parameter_value_mapping_from_dict({"value_type": value_type})
            else:
                raise RuntimeError(f"Unknown value type '{value_type}'")
        self.beginRemoveRows(list_index, 0, len(flattened_mappings.display_names) - 1)
        mapping_list.flattened_mappings = None
        self.endRemoveRows()
        flattened_mappings.set_value_components(value_mappings)
        self.beginInsertRows(list_index, 0, len(flattened_mappings.display_names) - 1)
        mapping_list.flattened_mappings = flattened_mappings
        self.endInsertRows()
        self.dataChanged.emit(list_index, list_index, [Role.FLATTENED_MAPPINGS])

    def set_import_entity_alternatives(self, table_row, list_row, import_entity_alternatives):
        """Adds or removes entity alternative imports.

        Args:
            table_row (int): source table row index
            list_row (int): mapping list row index
            import_entity_alternatives (bool): True to import entity alternatives, False to disable importing
        """
        table_index = self.index(table_row, 0)
        list_index = self.index(list_row, 0, table_index)
        mapping_list = self._mappings[table_row].mapping_list[list_row]
        flattened_mappings = mapping_list.flattened_mappings
        self.beginRemoveRows(list_index, 0, len(flattened_mappings.display_names) - 1)
        mapping_list.flattened_mappings = None
        self.endRemoveRows()
        flattened_mappings.set_import_entity_alternatives(import_entity_alternatives)
        self.beginInsertRows(list_index, 0, len(flattened_mappings.display_names) - 1)
        mapping_list.flattened_mappings = flattened_mappings
        self.endInsertRows()
        self.dataChanged.emit(list_index, list_index, [Role.FLATTENED_MAPPINGS])

    def change_component_mapping(self, flattened_mappings, index, new_type, new_ref):
        """
        Pushes :class:`SetComponentMappingType` to the undo stack.

        Args:
            flattened_mappings (FlattenedMappings): flattened mappings
            index (QModelIndex): flattened mappings index
            new_type (str): name of the new type
            new_ref (str or int): component mapping reference

        Returns:
            bool: True if the operation was successful, False otherwise
        """
        row = index.row()
        previous_type = flattened_mappings.display_position_type(row)
        previous_ref = flattened_mappings.display_position(row)
        parent_index = index.parent()
        mapping_list_row = parent_index.row()
        table_row = parent_index.parent().row()
        commands = []
        if new_type != previous_type:
            commands.append(
                SetMappingPositionType(table_row, mapping_list_row, row, self, new_type, previous_type, previous_ref)
            )
        if new_ref != previous_ref:
            commands.append(
                SetMappingPosition(
                    table_row, mapping_list_row, row, self, new_type, new_ref, previous_type, previous_ref
                )
            )
        if not commands:
            return False
        self._undo_stack.beginMacro("mapping type and reference change")
        for command in commands:
            self._undo_stack.push(command)
        self._undo_stack.endMacro()
        return True

    def set_mapping_position_type(self, table_row, list_row, row, position_type):
        """Modifies mapping component's position type.

        Args:
            table_row (int): source table row index
            list_row (int): mapping list row index
            row (int): mapping component row index
            position_type (str): new type
        """
        flattened_mappings = self._mappings[table_row].mapping_list[list_row].flattened_mappings
        flattened_mappings.set_display_position_type(row, position_type)
        table_index = self.index(table_row, 0)
        list_index = self.index(list_row, 0, table_index)
        top_left = self.index(row, 1, list_index)
        bottom_right = self.index(self.rowCount(list_index) - 1, 2, list_index)
        self.dataChanged.emit(
            top_left,
            bottom_right,
            [Qt.ItemDataRole.BackgroundRole, Qt.ItemDataRole.DisplayRole, Qt.ItemDataRole.ToolTipRole],
        )

    def set_mapping_position(self, table_row, list_row, row, position_type, position):
        """Modifies mapping component's position

        Args:
            table_row (int): source table row index
            list_row (int): mapping list row index
            row (int): mapping component row index
            position_type (str): new type
            position (str or int): new position
        """
        flattened_mappings = self._mappings[table_row].mapping_list[list_row].flattened_mappings
        flattened_mappings.set_display_position(row, position_type, position)
        table_index = self.index(table_row, 0)
        list_index = self.index(list_row, 0, table_index)
        top_left = self.index(row, 1, list_index)
        bottom_right = self.index(self.rowCount(list_index) - 1, 2, list_index)
        self.dataChanged.emit(
            top_left,
            bottom_right,
            [Qt.ItemDataRole.BackgroundRole, Qt.ItemDataRole.DisplayRole, Qt.ItemDataRole.ToolTipRole],
        )
        self._recommend_source_table_column_or_row_types(flattened_mappings, row)

    def set_import_entities(self, table_row, list_row, import_entities):
        """Sets the import entities flag.

        Args:
            table_row (int): source table row index
            list_row (int): mapping list row index
            import_entities (bool): flag value
        """
        flattened_mappings = self._mappings[table_row].mapping_list[list_row].flattened_mappings
        flattened_mappings.set_import_entities(import_entities)
        table_index = self.index(table_row, 0)
        list_index = self.index(list_row, 0, table_index)
        self.dataChanged.emit(list_index, list_index, [Role.FLATTENED_MAPPINGS])

    def set_read_start_row(self, table_row, list_row, start_row):
        """Sets mappings' read start row.

        Args:
            table_row (int): source table row index
            list_row (int): mapping list row index
            start_row (int): read start row
        """
        flattened_mappings = self._mappings[table_row].mapping_list[list_row].flattened_mappings
        flattened_mappings.set_read_start_row(start_row)
        table_index = self.index(table_row, 0)
        list_index = self.index(list_row, 0, table_index)
        self.dataChanged.emit(list_index, list_index, [Role.FLATTENED_MAPPINGS])

    def set_skip_columns(self, table_row, list_row, skip_columns):
        """Sets mappings' read start row.

        Args:
            table_row (int): source table row index
            list_row (int): mapping list row index
            skip_columns (list): skipped column indexes or names
        """
        if skip_columns is None:
            skip_columns = []
        flattened_mappings = self._mappings[table_row].mapping_list[list_row].flattened_mappings
        flattened_mappings.set_skip_columns(sorted(list(set(skip_columns))))
        table_index = self.index(table_row, 0)
        list_index = self.index(list_row, 0, table_index)
        self.dataChanged.emit(list_index, list_index, [Role.FLATTENED_MAPPINGS])

    def set_time_series_repeat_flag(self, table_row, list_row, repeat):
        """Sets the repeat time series flag.

        Args:
            table_row (int): source table row index
            list_row (int): mapping list row index
            repeat (bool): flag value
        """
        flattened_mappings = self._mappings[table_row].mapping_list[list_row].flattened_mappings
        if not flattened_mappings.is_time_series_value():
            return
        flattened_mappings.value_mapping().options["repeat"] = repeat
        table_index = self.index(table_row, 0)
        list_index = self.index(list_row, 0, table_index)
        self.dataChanged.emit(list_index, list_index, [Role.FLATTENED_MAPPINGS])

    def set_filter_re(self, table_row, list_row, row, filter_re):
        """Sets filter regular expression.

        Args:
            table_row (int): source table row index
            list_row (int): mapping list row index
            row (int): mapping component row index
            filter_re (str): filter regular expression
        """
        flattened_mappings = self._mappings[table_row].mapping_list[list_row].flattened_mappings
        flattened_mappings.component_at(row).filter_re = filter_re
        table_index = self.index(table_row, 0)
        list_index = self.index(list_row, 0, table_index)
        index = self.index(row, FlattenedColumn.REGEXP, list_index)
        self.dataChanged.emit(index, index, [Qt.ItemDataRole.DisplayRole])

    @staticmethod
    def check_validity_of_columns(list_index, header):
        """Checks that the mapping doesn't have column refs
        that are larger than the source table column count

        Args:
            list_index (QModelIndex): index to mappings list
            header (Iterable): source table header
        Returns:
            bool: True if no source ref. is out of range, False otherwise
        """
        mapping_list_item = list_index.internalPointer()
        table_name = mapping_list_item.source_table_item.name
        return mapping_list_item.flattened_mappings.root_mapping.check_for_invalid_column_refs(header, table_name)

    @staticmethod
    def polish_mapping(list_index, header):
        """Polishes flattened mappings.

        Args:
            list_index (QModelIndex): index to mappings list
            header (Iterable): source table header
        """
        mapping_list_item = list_index.internalPointer()
        table_name = mapping_list_item.source_table_item.name
        mapping_name = mapping_list_item.name
        mapping_list_item.flattened_mappings.root_mapping.polish(table_name, header, mapping_name, for_preview=True)
        # We don't emit dataChanged here as polishing is just beautification and
        # it would mess up the undo system which relies on real changes.

    def _recommend_source_table_column_or_row_types(self, flattened_mappings, row):
        """Checks if source table conversion specifications need an update.

        Args:
            flattened_mappings (FlattenedMappings): flattened mappings
            row (int): display row index
        """
        self._recommend_float_type_for_values(flattened_mappings)
        display_name = flattened_mappings.display_names[row]
        if display_name.endswith("indexes") and flattened_mappings.is_time_series_value():
            self._recommend_date_time_type(flattened_mappings.component_at(row))
        elif not display_name.endswith("values"):
            self._recommend_string_type(flattened_mappings.component_at(row))

    def _recommend_float_type_for_values(self, flattened_mappings):
        """

        Args:
            flattened_mappings (FlattenedMappings): flattened mappings
        """
        if not flattened_mappings.has_value_component():
            return
        visual_row = next(
            (k for k, name in enumerate(flattened_mappings.display_names) if name.endswith("values")), None
        )
        if visual_row is None:
            return
        component = flattened_mappings.component_at(visual_row)
        if not flattened_mappings.root_mapping.is_pivoted():
            self._recommend_float_type(component)
            return
        not_pivoted_columns = set(flattened_mappings.root_mapping.non_pivoted_columns())
        ignored_columns = set(flattened_mappings.skip_columns())
        excluded_columns = sorted(list(not_pivoted_columns | ignored_columns))
        self.multi_column_type_recommended.emit(excluded_columns, FloatConvertSpec())

    def _recommend_string_type(self, component):
        """Signals a need to change source table conversion specification to StringConvertSpec.

        Args:
            component (ImportMapping): mapping component
        """
        self._recommend_type(component, StringConvertSpec())

    def _recommend_float_type(self, component):
        """Signals a need to change source table conversion specification to FloatConvertSpec.

        Args:
            component (ImportMapping): mapping component
        """
        self._recommend_type(component, FloatConvertSpec())

    def _recommend_date_time_type(self, component):
        """Signals a need to change source table conversion specification to DateTimeConvertSpec.

        Args:
            component (ImportMapping): mapping component
        """
        self._recommend_type(component, DateTimeConvertSpec())

    def _recommend_type(self, component, convert_spec):
        """Signals a need to change source table conversion specifications.

        Args:
            component (ImportMapping): mapping component
            convert_spec (ConvertSpec): row/column conversion specification
        """
        if not isinstance(component.position, int):
            return
        if component.position >= 0:
            self.row_or_column_type_recommended.emit(component.position, convert_spec, Qt.Orientation.Horizontal)
        else:
            self.row_or_column_type_recommended.emit(-(component.position + 1), convert_spec, Qt.Orientation.Vertical)

    def set_tables_editable(self, editable):
        """Enables or disables table name editing.

        Args:
            editable (bool): True to enable table name editing, False to disable
        """
        for item in self._mappings[1:]:
            if item.empty:
                continue
            item.editable = editable

    def add_empty_row(self):
        """Adds the special 'unnamed table' row at the end of table list."""
        last_row = len(self._mappings) - 1
        empty_item = SourceTableItem(
            UNNAMED_TABLE_NAME,
            checked=False,
            checkable=False,
            editable=True,
            accepts_drags_and_drops=False,
            real=False,
            empty=True,
        )
        self.beginInsertRows(QModelIndex(), last_row, last_row)
        self._mappings.append(empty_item)
        self.endInsertRows()

    def insertRows(self, row, count, parent=QModelIndex()):
        parent_item = parent.internalPointer()
        if isinstance(parent_item, SourceTableItem) and parent_item.real:
            self.beginInsertRows(parent, row, row + count - 1)
            for _ in range(count):
                list_item = MappingListItem(self._unique_mapping_name(parent_item))
                list_item.set_flattened_mappings(FlattenedMappings(self.create_default_mapping()))
                parent_item.insert_to_mapping_list(row, list_item)
            self.endInsertRows()
            return True
        return False

    def moveRows(self, sourceParent, sourceRow, count, destinationParent, destinationChild):

        source_parent_item = sourceParent.internalPointer()
        if source_parent_item is None:
            return self._move_source_table_list_rows(sourceRow, count, destinationChild)
        if isinstance(source_parent_item, SourceTableItem):
            return self._move_mapping_list_rows(sourceParent, sourceRow, count, destinationParent, destinationChild)
        return False

    def _move_source_table_list_rows(self, source_row, count, destination_child):
        """Moves row in mapping list.

        Args:
            source_row (int): source row index
            count (int): mapping count
            destination_child (int): destination row index

        Returns:
            bool: True if operation was successful, False otherwise
        """
        items = self._mappings[source_row : source_row + count]
        is_legal_move = self.beginMoveRows(
            QModelIndex(), source_row, source_row + count - 1, QModelIndex(), destination_child
        )
        if not is_legal_move:
            return False
        self._mappings = self._mappings[:source_row] + self._mappings[source_row + count :]
        if source_row < destination_child:
            destination_child -= count
        for i, item in enumerate(items):
            self._mappings.insert(destination_child + i, item)
        self.endMoveRows()
        return True

    def _move_mapping_list_rows(self, source_parent, source_row, count, destination_parent, destination_child):
        """Moves row in mapping list.

        Args:
            source_parent (QModelIndex): source parent
            source_row (int): source row index
            count (int): mapping count
            destination_parent (QModelIndex): destination parent
            destination_child (int): destination row index

        Returns:
            bool: True if operation was successful, False otherwise
        """
        source_table_item = source_parent.internalPointer()
        destination_table_item = destination_parent.internalPointer()
        if not isinstance(destination_table_item, SourceTableItem):
            return False
        is_internal_move = source_table_item is destination_table_item
        source_list = source_table_item.mapping_list
        items = source_list[source_row : source_row + count]
        is_legal_move = self.beginMoveRows(
            source_parent, source_row, source_row + count - 1, destination_parent, destination_child
        )
        if not is_legal_move:
            return False
        source_table_item.mapping_list = source_list[:source_row] + source_list[source_row + count :]
        if is_internal_move and source_row < destination_child:
            destination_child -= count
        for i, item in enumerate(items):
            destination_table_item.insert_to_mapping_list(destination_child + i, item)
        self.endMoveRows()
        return True

    def removeRows(self, row, count, parent=QModelIndex()):
        parent_item = parent.internalPointer()
        if parent_item is None:
            self.beginRemoveRows(parent, row, row + count - 1)
            self._mappings = self._mappings[:row] + self._mappings[row + count :]
            self.endRemoveRows()
            return True
        if isinstance(parent_item, SourceTableItem):
            self.beginRemoveRows(parent, row, row + count - 1)
            parent_item.mapping_list = parent_item.mapping_list[:row] + parent_item.mapping_list[row + count :]
            self.endRemoveRows()
            return True
        return False

    @staticmethod
    def create_default_mapping():
        """Creates a default object class mapping.

        Returns:
            ImportMapping: new mapping root
        """
        return default_import_mapping("EntityClass")

    def has_mapping_name(self, table_row, name):
        """Checks if a name exists in mapping list.

        Args:
            table_row (int): source table row index
            name (str): name to check

        Returns:
            bool: True if name exists, False otherwise
        """
        return name in {i.name for i in self._mappings[table_row].mapping_list}

    def unique_mapping_name(self, table_row, prefix):
        """Generates a unique mapping name.

        Args:
            table_row (int): source table row index
            prefix (str): name prefix

        Returns:
            str: name
        """
        return self._unique_mapping_name(self._mappings[table_row], prefix)

    @staticmethod
    def _unique_mapping_name(table_item, prefix="Mapping"):
        """Generates a unique mapping name.

        Args:
            table_item (SourceTableItem): source table item
            prefix (str): name prefix

        Returns:
            str: name
        """
        return unique_name(prefix, [i.name for i in table_item.mapping_list])

    def set_source_table_items_into_specification(self):
        """Sets the in_specification flag for all appropriate source table items."""
        bottom = 0
        for mapping in self._mappings[1:]:
            mapping.in_specification = True
            bottom += 1
        if bottom > 0:
            top_left = self.index(1, 0)
            bottom_right = self.index(bottom, 0)
            self.dataChanged.emit(top_left, bottom_right, [Qt.ItemDataRole.DisplayRole])

    def remove_tables_not_in_source_and_specification(self):
        """Removes source table items that are not in source or saved in specification."""
        start = 1
        for row, table_item in reversed(list(enumerate(self._mappings[start:], start))):
            if not table_item.in_source and not table_item.in_specification:
                self.removeRow(row)

    def cross_check_source_table_names(self, table_names):
        """Sets the in_source flag for all appropriate source table items.

        Args:
            table_names (set of str): existing source table names
        """
        changed_rows = []
        start = 1
        for row, table_item in enumerate(self._mappings[start:], start):
            in_source = table_item.name in table_names
            changed = False
            if table_item.in_source != in_source:
                table_item.in_source = in_source
                changed = True
            if table_item.real != in_source:
                table_item.real = in_source
                changed = True
            if changed:
                changed_rows.append(row)
        if changed_rows:
            top_left = self.index(changed_rows[0], 0)
            bottom_right = self.index(changed_rows[-1], 0)
            self.dataChanged.emit(top_left, bottom_right, [Qt.ItemDataRole.ForegroundRole])



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Exporter's execute kernel (do_work), as target for a multiprocess.Process"""
from datetime import datetime
import os
from pathlib import Path
from time import time
from spine_engine.utils.helpers import write_filter_id_file
from spinedb_api import DatabaseMapping, SpineDBAPIError
from spinedb_api.spine_io.exporters.csv_writer import CsvWriter
from spinedb_api.spine_io.exporters.excel_writer import ExcelWriter
from spinedb_api.spine_io.exporters.gdx_writer import GdxWriter
from spinedb_api.spine_io.exporters.sql_writer import SqlWriter
from spinedb_api.spine_io.exporters.writer import WriterException, write
from ..utils import convert_to_sqlalchemy_url, split_url_credentials
from .specification import OutputFormat, Specification


def do_work(
    process,
    specification,
    output_time_stamps,
    cancel_on_error,
    gams_path,
    out_dir,
    databases,
    out_urls,
    filter_id,
    filter_subdirectory,
    logger,
):
    """
    Exports databases using given specification as export mapping.

    Args:
        process (Process): unused
        specification (dict): export specification dictionary
        output_time_stamps (bool): if True, puts output files into time stamped subdirectories
        cancel_on_error (bool): if True, bails out on non-fatal errors
        gams_path (str): path to GAMS installation
        out_dir (str): base output directory
        databases (dict): databases to export
        out_urls (dict): output URLs
        filter_id (str): filter id
        filter_subdirectory (str): name of extra subdirectory used when filters have been applied
        logger (LoggerInterface): a logger

    Returns:
        tuple: boolean success flag, dictionary of output files
    """
    specification = Specification.from_dict(specification)
    successes = []
    written_files = {}
    for url, output_label in databases.items():
        try:
            database_map = DatabaseMapping(url)
        except SpineDBAPIError as error:
            sanitized_url, _ = split_url_credentials(url)
            logger.msg_error.emit(f"Failed to export <b>{sanitized_url}</b>: {error}")
            if cancel_on_error:
                return False, written_files
            successes.append(False)
            continue
        try:
            out_url = out_urls.get(url)
            if specification.output_format == OutputFormat.SQL and out_url is not None:
                successful = _export_to_database(
                    database_map, specification, out_url, successes, cancel_on_error, logger
                )
            else:
                successful = _export_to_file(
                    database_map,
                    specification,
                    output_label,
                    output_time_stamps,
                    successes,
                    written_files,
                    cancel_on_error,
                    gams_path,
                    out_dir,
                    filter_id,
                    filter_subdirectory,
                    logger,
                )
            if not successful:
                return False, written_files
        finally:
            database_map.close()
    return all(successes), written_files


def _export_to_file(
    database_map,
    specification,
    output_label,
    output_time_stamps,
    successes,
    written_files,
    cancel_on_error,
    gams_path,
    out_dir,
    filter_id,
    filter_subdirectory,
    logger,
):
    """Exports into file(s) including a new SQLite file.

    Args:
        database_map (DatabaseMapping): source database map
        specification (Specification): export specification dictionary
        output_label (str): output label
        output_time_stamps (bool): if True, puts output files into time stamped subdirectories
        successes (list of bool): history of success statuses
        written_files (dict): mapping from output label to completed output files
        cancel_on_error (bool): if True, bails out on non-fatal errors
        gams_path (str): path to GAMS installation
        out_dir (str): base output directory
        filter_id (str): filter id
        filter_subdirectory (str): name of extra subdirectory used when filters have been applied
        logger (LoggerInterface): a logger

    Returns:
        bool: True if operation was successful, False otherwise
    """
    output_file_name = _add_extension(output_label, specification.output_format)
    out_path = _subdirectory_for_fork(output_file_name, out_dir, output_time_stamps, filter_subdirectory)
    try:
        file = Path(out_path)
        file.parent.mkdir(parents=True, exist_ok=True)
        if file.exists():
            file.unlink()
        writer = make_writer(specification.output_format, out_path, gams_path)
        specifications = specification.enabled_specifications().values()
        mappings = (m.root for m in specifications)
        header_always = (m.always_export_header for m in specifications)
        group_fns = (m.group_fn for m in specifications)
        write(database_map, writer, *mappings, empty_data_header=header_always, group_fns=group_fns)
    except (FileNotFoundError, PermissionError, WriterException) as e:
        logger.msg_error.emit(str(e))
        if cancel_on_error:
            return False
        successes.append(False)
    else:
        if isinstance(writer, CsvWriter):
            files = writer.output_files()
        else:
            files = {out_path}
        written_files[output_label] = files
        if len(files) > 1:
            anchors = []
            for path in (Path(f) for f in files):
                anchors.append(f"\t<a style='color:#BB99FF;' title='{path}' href='file:///{path}'>{path.name}</a>")
            logger.msg_success.emit(f"Wrote multiple files:<br>{'<br>'.join(anchors)}")
        else:
            only_file = Path(next(iter(files)))
            file_anchor = (
                f"<a style='color:#BB99FF;' title='{only_file}' href='file:///{only_file}'>{only_file.name}</a>"
            )
            logger.msg_success.emit(f"Wrote {file_anchor}")
        if filter_id:
            write_filter_id_file(filter_id, Path(out_path).parent)
        successes.append(True)
    return True


def _export_to_database(database_map, specification, out_url, successes, cancel_on_error, logger):
    """Exports into fixed output database.

    Args:
        database_map (DatabaseMapping): source database map
        specification (Specification): export specification dictionary
        out_url (dict): output URL
        successes (list of bool): history of success statuses
        cancel_on_error (bool): if True, bails out on non-fatal errors
        logger (LoggerInterface): a logger

    Returns:
        bool: True if operation was successful, False otherwise
    """
    url = convert_to_sqlalchemy_url(out_url, logger=logger)
    if url is None:
        return True
    try:
        writer = SqlWriter(str(url), overwrite_existing=False)
        specifications = specification.enabled_specifications().values()
        mappings = (m.root for m in specifications)
        header_always = (m.always_export_header for m in specifications)
        group_fns = (m.group_fn for m in specifications)
        write(database_map, writer, *mappings, empty_data_header=header_always, group_fns=group_fns)
    except WriterException as e:
        logger.msg_error.emit(str(e))
        if cancel_on_error:
            return False
        successes.append(False)
    else:
        logger.msg_success.emit("Wrote to database.")
        successes.append(True)
    return True


def make_writer(output_format, out_path, gams_path):
    """
    Constructs a writer.

    Args:
        output_format (OutputFormat): output format
        out_path (str): path to output file
        gams_path (str): path to GAMS installation

    Returns:
        Writer: a writer
    """
    if output_format == OutputFormat.CSV:
        path = Path(out_path)
        return CsvWriter(path.parent, path.name)
    if output_format == OutputFormat.EXCEL:
        return ExcelWriter(out_path)
    if output_format == OutputFormat.SQL:
        return SqlWriter(out_path, overwrite_existing=True)
    return GdxWriter(out_path, gams_path)


def _add_extension(label, file_format):
    """Adds file format dependent extension to ``label`` if it is missing one.

    Args:
        label (str): label to add the extension to
        file_format (OutputFormat): file format

    Returns:
        str: file name
    """
    name, _, label_extension = label.rpartition(".")
    if name and file_format.is_compatible_file_extension(label_extension):
        return label
    return label + "." + file_format.file_extension()


def _subdirectory_for_fork(output_file_name, data_dir, output_time_stamps, filter_id_hash):
    """
    Creates scenario/tool based output directory for forked workflow.

    Args:
        output_file_name (str): file name
        data_dir (str): project item's data directory
        output_time_stamps (bool): True if time stamp data should be included in the output path
        filter_id_hash (str): hashed filter id

    Returns:
        str: absolute output path
    """
    if output_time_stamps:
        stamp = datetime.fromtimestamp(time())
        time_stamp = "run@" + stamp.isoformat(timespec="seconds").replace(":", ".")
    else:
        time_stamp = ""
    if filter_id_hash:
        if time_stamp:
            path = os.path.join(data_dir, filter_id_hash + "_" + time_stamp, output_file_name)
        else:
            path = os.path.join(data_dir, filter_id_hash, output_file_name)
    else:
        path = os.path.join(data_dir, time_stamp, output_file_name)
    return path



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################
import unittest
import numpy
from spine_items.exporter.preview_table_writer import TableWriter


class TestTableWriter(unittest.TestCase):
    def setUp(self):
        self._writer = TableWriter()
        self._writer.start()

    def tearDown(self):
        self._writer.finish()

    def test_write_row_float64_converted_to_float(self):
        self.assertTrue(self._writer.start_table("My table", {}))
        self.assertTrue(self._writer.write_row([numpy.float64(2.3)]))
        self._writer.finish_table()
        tables = self._writer.tables
        self.assertEqual(tables, {"My table": [[2.3]]})


if __name__ == "__main__":
    unittest.main()



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Unit tests for ``flattened_mappings`` module."""
import unittest
from spine_items.importer.flattened_mappings import FlattenedMappings, MappingType
from spinedb_api import import_mapping_from_dict
from spinedb_api.import_mapping.import_mapping import default_import_mapping
from spinedb_api.import_mapping.import_mapping_compat import parameter_mapping_from_dict


class TestFlattenedMappings(unittest.TestCase):
    def test_increasing_relationship_dimensions_keeps_import_objects_flags_consistent(self):
        mapping_dicts = [
            {"map_type": "RelationshipClass", "position": "hidden", "value": "o_to_q"},
            {"map_type": "RelationshipClassObjectClass", "position": "hidden", "value": "o"},
            {"map_type": "Relationship", "position": "hidden", "value": "relationship"},
            {"map_type": "RelationshipObject", "position": -1, "import_objects": True},
            {"map_type": "RelationshipMetadata", "position": "hidden"},
            {"map_type": "ParameterDefinition", "position": "hidden", "value": "param"},
            {"map_type": "Alternative", "position": "hidden", "value": "base"},
            {"map_type": "ParameterValueMetadata", "position": "hidden"},
            {"map_type": "ParameterValueType", "position": "hidden", "value": "map"},
            {"map_type": "IndexName", "position": "hidden", "value": "time"},
            {"map_type": "ParameterValueIndex", "position": 0},
            {"map_type": "ExpandedValue", "position": "hidden"},
        ]
        root_mapping = import_mapping_from_dict(mapping_dicts)
        flattened_mappings = FlattenedMappings(root_mapping)
        self.assertTrue(flattened_mappings.import_entities())
        flattened_mappings.set_dimension_count(2)
        self.assertTrue(flattened_mappings.import_entities())

    def test_default_entity_class_mapping(self):
        root_mapping = default_import_mapping("EntityClass")
        flattened_mappings = FlattenedMappings(root_mapping)
        self.assertEqual(flattened_mappings.map_type, MappingType.EntityClass)
        self.assertEqual(flattened_mappings.display_names, ["Entity class names", "Entity names", "Entity metadata"])
        self.assertIs(flattened_mappings.root_mapping, root_mapping)
        self.assertIsNone(flattened_mappings.value_type)
        self.assertFalse(flattened_mappings.is_time_series_value())
        self.assertFalse(flattened_mappings.is_map_value())
        self.assertEqual(flattened_mappings.map_dimension_count(), 1)
        self.assertEqual(flattened_mappings.read_start_row(), 0)
        self.assertEqual(flattened_mappings.skip_columns(), [])
        for row in range(len(flattened_mappings.display_names)):
            with self.subTest(row=row):
                self.assertEqual(flattened_mappings.display_position_type(row), "None")
                self.assertIsNone(flattened_mappings.display_position(row))
                self.assertEqual(flattened_mappings.display_row_issues(row), [])
        self.assertTrue(flattened_mappings.can_have_dimensions())
        self.assertFalse(flattened_mappings.has_dimensions())
        self.assertEqual(flattened_mappings.dimension_count(), 0)
        self.assertTrue(flattened_mappings.may_import_entity_alternatives())
        self.assertFalse(flattened_mappings.import_entity_alternatives())
        self.assertFalse(flattened_mappings.may_import_entities())
        self.assertFalse(flattened_mappings.import_entities())
        self.assertTrue(flattened_mappings.has_parameters())
        self.assertEqual(flattened_mappings.display_parameter_type(), "None")
        self.assertFalse(flattened_mappings.has_value_component())

    def test_add_dimensions_to_entity_class_mapping(self):
        root_mapping = default_import_mapping("EntityClass")
        flattened_mappings = FlattenedMappings(root_mapping)
        self.assertFalse(flattened_mappings.has_dimensions())
        self.assertEqual(flattened_mappings.dimension_count(), 0)
        flattened_mappings.set_dimension_count(1)
        self.assertTrue(flattened_mappings.has_dimensions())
        self.assertEqual(flattened_mappings.dimension_count(), 1)
        self.assertEqual(
            flattened_mappings.display_names,
            ["Entity class names", "Dimension names", "Element names", "Entity metadata"],
        )
        self.assertTrue(flattened_mappings.may_import_entities())
        self.assertFalse(flattened_mappings.import_entities())
        flattened_mappings.set_dimension_count(2)
        self.assertTrue(flattened_mappings.has_dimensions())
        self.assertEqual(flattened_mappings.dimension_count(), 2)
        self.assertEqual(
            flattened_mappings.display_names,
            [
                "Entity class names",
                "Dimension names 1",
                "Dimension names 2",
                "Element names 1",
                "Element names 2",
                "Entity metadata",
            ],
        )
        self.assertTrue(flattened_mappings.may_import_entities())
        self.assertFalse(flattened_mappings.import_entities())
        flattened_mappings.set_dimension_count(0)
        self.assertFalse(flattened_mappings.has_dimensions())
        self.assertEqual(flattened_mappings.dimension_count(), 0)
        self.assertEqual(flattened_mappings.display_names, ["Entity class names", "Entity names", "Entity metadata"])
        self.assertFalse(flattened_mappings.may_import_entities())
        self.assertFalse(flattened_mappings.import_entities())

    def test_add_parameter_definition_to_entity_class_mapping(self):
        root_mapping = default_import_mapping("EntityClass")
        flattened_mappings = FlattenedMappings(root_mapping)
        self.assertEqual(flattened_mappings.display_parameter_type(), "None")
        self.assertFalse(flattened_mappings.has_value_component())
        self.assertEqual(flattened_mappings.value_type_label(), "<no label>")
        parameter_root = parameter_mapping_from_dict({"map_type": "ParameterDefinition"})
        flattened_mappings.set_parameter_components(parameter_root)
        self.assertEqual(flattened_mappings.display_parameter_type(), "Definition")
        self.assertTrue(flattened_mappings.has_value_component())
        self.assertEqual(flattened_mappings.value_type_label(), "Default value:")
        self.assertEqual(
            flattened_mappings.display_names,
            [
                "Entity class names",
                "Entity names",
                "Entity metadata",
                "Parameter names",
                "Value list names",
                "Parameter default values",
            ],
        )

    def test_remove_parameter_definition_from_entity_class_mapping(self):
        root_mapping = default_import_mapping("EntityClass")
        flattened_mappings = FlattenedMappings(root_mapping)
        parameter_root = parameter_mapping_from_dict({"map_type": "ParameterDefinition"})
        flattened_mappings.set_parameter_components(parameter_root)
        flattened_mappings.set_parameter_components(None)
        self.assertEqual(flattened_mappings.display_parameter_type(), "None")
        self.assertFalse(flattened_mappings.has_value_component())
        self.assertEqual(flattened_mappings.value_type_label(), "<no label>")
        self.assertEqual(
            flattened_mappings.display_names,
            [
                "Entity class names",
                "Entity names",
                "Entity metadata",
            ],
        )

    def test_add_parameter_value_to_entity_class_mapping(self):
        root_mapping = default_import_mapping("EntityClass")
        flattened_mappings = FlattenedMappings(root_mapping)
        self.assertEqual(flattened_mappings.display_parameter_type(), "None")
        self.assertFalse(flattened_mappings.has_value_component())
        self.assertEqual(flattened_mappings.value_type_label(), "<no label>")
        parameter_root = parameter_mapping_from_dict({"map_type": "ParameterValue"})
        flattened_mappings.set_parameter_components(parameter_root)
        self.assertEqual(flattened_mappings.display_parameter_type(), "Value")
        self.assertTrue(flattened_mappings.has_value_component())
        self.assertEqual(flattened_mappings.value_type_label(), "Value:")
        self.assertEqual(
            flattened_mappings.display_names,
            [
                "Entity class names",
                "Entity names",
                "Entity metadata",
                "Alternative names",
                "Parameter names",
                "Parameter value metadata",
                "Parameter values",
            ],
        )

    def test_remove_parameter_value_from_entity_class_mapping(self):
        root_mapping = default_import_mapping("EntityClass")
        flattened_mappings = FlattenedMappings(root_mapping)
        parameter_root = parameter_mapping_from_dict({"map_type": "ParameterValue"})
        flattened_mappings.set_parameter_components(parameter_root)
        flattened_mappings.set_parameter_components(None)
        self.assertEqual(flattened_mappings.display_parameter_type(), "None")
        self.assertFalse(flattened_mappings.has_value_component())
        self.assertEqual(flattened_mappings.value_type_label(), "<no label>")
        self.assertEqual(
            flattened_mappings.display_names,
            [
                "Entity class names",
                "Entity names",
                "Entity metadata",
            ],
        )

    def test_add_entity_alternative_to_entity_class_mapping(self):
        root_mapping = default_import_mapping("EntityClass")
        flattened_mappings = FlattenedMappings(root_mapping)
        self.assertFalse(flattened_mappings.import_entity_alternatives())
        flattened_mappings.set_import_entity_alternatives(True)
        self.assertTrue(flattened_mappings.import_entity_alternatives())
        self.assertEqual(
            flattened_mappings.display_names,
            [
                "Entity class names",
                "Entity names",
                "Entity metadata",
                "Alternative names",
                "Entity activities",
            ],
        )

    def test_remove_entity_alternative_from_entity_class_mapping(self):
        root_mapping = default_import_mapping("EntityClass")
        flattened_mappings = FlattenedMappings(root_mapping)
        flattened_mappings.set_import_entity_alternatives(True)
        flattened_mappings.set_import_entity_alternatives(False)
        self.assertFalse(flattened_mappings.import_entity_alternatives())
        self.assertEqual(
            flattened_mappings.display_names,
            [
                "Entity class names",
                "Entity names",
                "Entity metadata",
            ],
        )

    def test_add_entity_alternative_and_parameter_value_to_entity_class_mapping(self):
        root_mapping = default_import_mapping("EntityClass")
        flattened_mappings = FlattenedMappings(root_mapping)
        flattened_mappings.set_import_entity_alternatives(True)
        parameter_root = parameter_mapping_from_dict({"map_type": "ParameterValue"})
        flattened_mappings.set_parameter_components(parameter_root)
        self.assertTrue(flattened_mappings.import_entity_alternatives())
        self.assertEqual(flattened_mappings.display_parameter_type(), "Value")
        self.assertTrue(flattened_mappings.has_value_component())
        self.assertEqual(flattened_mappings.value_type_label(), "Value:")
        self.assertEqual(
            flattened_mappings.display_names,
            [
                "Entity class names",
                "Entity names",
                "Entity metadata",
                "Alternative names",
                "Entity activities",
                "Parameter names",
                "Parameter value metadata",
                "Parameter values",
            ],
        )

    def test_add_parameter_value_and_entity_alternative_to_entity_class_mapping(self):
        root_mapping = default_import_mapping("EntityClass")
        flattened_mappings = FlattenedMappings(root_mapping)
        parameter_root = parameter_mapping_from_dict({"map_type": "ParameterValue"})
        flattened_mappings.set_parameter_components(parameter_root)
        flattened_mappings.set_import_entity_alternatives(True)
        self.assertTrue(flattened_mappings.import_entity_alternatives())
        self.assertEqual(flattened_mappings.display_parameter_type(), "Value")
        self.assertTrue(flattened_mappings.has_value_component())
        self.assertEqual(flattened_mappings.value_type_label(), "Value:")
        self.assertEqual(
            flattened_mappings.display_names,
            [
                "Entity class names",
                "Entity names",
                "Entity metadata",
                "Alternative names",
                "Entity activities",
                "Parameter names",
                "Parameter value metadata",
                "Parameter values",
            ],
        )

    def test_remove_parameter_value_from_entity_class_mapping_with_entity_alternative(self):
        root_mapping = default_import_mapping("EntityClass")
        flattened_mappings = FlattenedMappings(root_mapping)
        parameter_root = parameter_mapping_from_dict({"map_type": "ParameterValue"})
        flattened_mappings.set_parameter_components(parameter_root)
        flattened_mappings.set_import_entity_alternatives(True)
        flattened_mappings.set_parameter_components(None)
        self.assertTrue(flattened_mappings.import_entity_alternatives())
        self.assertEqual(flattened_mappings.display_parameter_type(), "None")
        self.assertFalse(flattened_mappings.has_value_component())
        self.assertEqual(flattened_mappings.value_type_label(), "<no label>")
        self.assertEqual(
            flattened_mappings.display_names,
            [
                "Entity class names",
                "Entity names",
                "Entity metadata",
                "Alternative names",
                "Entity activities",
            ],
        )

    def test_remove_entity_alternative_from_entity_class_mapping_with_parameter_value(self):
        root_mapping = default_import_mapping("EntityClass")
        flattened_mappings = FlattenedMappings(root_mapping)
        parameter_root = parameter_mapping_from_dict({"map_type": "ParameterValue"})
        flattened_mappings.set_parameter_components(parameter_root)
        flattened_mappings.set_import_entity_alternatives(True)
        flattened_mappings.set_import_entity_alternatives(False)
        self.assertFalse(flattened_mappings.import_entity_alternatives())
        self.assertEqual(flattened_mappings.display_parameter_type(), "Value")
        self.assertTrue(flattened_mappings.has_value_component())
        self.assertEqual(flattened_mappings.value_type_label(), "Value:")
        self.assertEqual(
            flattened_mappings.display_names,
            [
                "Entity class names",
                "Entity names",
                "Entity metadata",
                "Alternative names",
                "Parameter names",
                "Parameter value metadata",
                "Parameter values",
            ],
        )

    def test_remove_entity_alternative_and_parameter_value_from_entity_class_mapping(self):
        root_mapping = default_import_mapping("EntityClass")
        flattened_mappings = FlattenedMappings(root_mapping)
        parameter_root = parameter_mapping_from_dict({"map_type": "ParameterValue"})
        flattened_mappings.set_parameter_components(parameter_root)
        flattened_mappings.set_import_entity_alternatives(True)
        flattened_mappings.set_import_entity_alternatives(False)
        flattened_mappings.set_parameter_components(None)
        self.assertFalse(flattened_mappings.import_entity_alternatives())
        self.assertEqual(flattened_mappings.display_parameter_type(), "None")
        self.assertFalse(flattened_mappings.has_value_component())
        self.assertEqual(flattened_mappings.value_type_label(), "<no label>")
        self.assertEqual(
            flattened_mappings.display_names,
            [
                "Entity class names",
                "Entity names",
                "Entity metadata",
            ],
        )

    def test_remove_parameter_value_and_entity_alternative_from_entity_class_mapping(self):
        root_mapping = default_import_mapping("EntityClass")
        flattened_mappings = FlattenedMappings(root_mapping)
        parameter_root = parameter_mapping_from_dict({"map_type": "ParameterValue"})
        flattened_mappings.set_parameter_components(parameter_root)
        flattened_mappings.set_import_entity_alternatives(True)
        flattened_mappings.set_parameter_components(None)
        flattened_mappings.set_import_entity_alternatives(False)
        self.assertFalse(flattened_mappings.import_entity_alternatives())
        self.assertEqual(flattened_mappings.display_parameter_type(), "None")
        self.assertFalse(flattened_mappings.has_value_component())
        self.assertEqual(flattened_mappings.value_type_label(), "<no label>")
        self.assertEqual(
            flattened_mappings.display_names,
            [
                "Entity class names",
                "Entity names",
                "Entity metadata",
            ],
        )


if __name__ == "__main__":
    unittest.main()



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains Data transformer's specification."""
from spine_engine.project_item.project_item_specification import ProjectItemSpecification
from .item_info import ItemInfo
from .settings import EntityClassRenamingSettings, settings_from_dict


class DataTransformerSpecification(ProjectItemSpecification):
    """
    Data transformer's specification.

    Attributes:
        settings (FilterSettings): transformer's filter settings
    """

    def __init__(self, name, settings=None, description=None):
        """
        Args:
            name (str): specification's name
            settings (FilterSettings, optional): filter settings
            description (str, optional): specification's description
        """
        super().__init__(name, description, ItemInfo.item_type())
        self.settings = settings

    def is_equivalent(self, other):
        """See base class."""
        return self.name == other.name and self.description == other.description and self.settings == other.settings

    def to_dict(self):
        """See base class."""
        filter_dict = (
            {"type": self.settings.type(), "settings": self.settings.to_dict()} if self.settings is not None else None
        )
        return {
            "name": self.name,
            "item_type": ItemInfo.item_type(),
            "description": self.description,
            "filter": filter_dict,
        }

    @staticmethod
    def from_dict(specification_dict, logger):
        """
        Restores :class:`DataTransformerSpecification` from a dictionary.

        Args:
            specification_dict (dict): serialized specification
            logger (LoggerInterface): a logger

        Returns:
            DataTransformerSpecification: deserialized specification
        """
        name = specification_dict["name"]
        description = specification_dict["description"]
        filter_ = specification_dict.get("filter")
        if filter_ is not None:
            settings = settings_from_dict(filter_)
        else:
            # For legacy JSON.
            renaming = specification_dict.get("entity_class_name_map")
            if renaming is not None:
                settings = EntityClassRenamingSettings(renaming)
            else:
                settings = None
        return DataTransformerSpecification(name, settings, description)
